class Block implements Cloneable {
  PVector location, block_size, first_moving_location, second_moving_location, movement_axis, movement_constraint_min, movement_constraint_max;
  boolean collidable, must_update, is_img, visible, moving_to_second;
  String img_name;
  color rectColor;
  int[] block_rounding;
  PVector[] vertices;
  //float power, full_cooldown; //for spring block, shooter block
  JSONObject properties;
  
  Block() {
    must_update = false;
    visible = true;
    properties = parseJSONObject("{\"power\": 30, \"full_cooldown\": 120, \"cooldown\": 0, \"cactus_transition\": 0}");
    //block_rounding = getBlockRounding();
  }
  
  void update() {
    move();
  }
  
  void display() {
    if (!visible) {
      return;
    }
    if (is_img) {
      image(gt(img_name), location.x, location.y);
    } else {
      if (block_rounding == null) {
        block_rounding = getBlockRounding();
      }
      fill(rectColor);
      rect(location.x, location.y, block_size.x, block_size.y, block_rounding[0], block_rounding[1], block_rounding[2], block_rounding[3]);
    }
  }
  
  void onCollide() {}
  
  boolean onEnemyCollide(Enemy enemy) {
    return true;
  }
  
  void setMoving(float x1, float y1, float x2, float y2, float moving_speed) {
    first_moving_location = new PVector(x1, y1);
    second_moving_location = new PVector(x2, y2);
    must_update = true;
    moving_to_second = true;
    movement_axis = (new PVector(x2-x1, y2-y1)).setMag(moving_speed);
    movement_constraint_min = new PVector(min(first_moving_location.x, second_moving_location.x), min(first_moving_location.y, second_moving_location.y));
    movement_constraint_max = new PVector(max(first_moving_location.x, second_moving_location.x), max(first_moving_location.y, second_moving_location.y));
  }
  
  void move() {
    if (movement_axis != null) {
      //PVector plocation = new PVector(location.x, location.y);
      if (moving_to_second) {
        location.add(movement_axis);
      } else {
        location.sub(movement_axis);
      }
      
      if (location.x != constrain(location.x, movement_constraint_min.x, movement_constraint_max.x) || location.y != constrain(location.y, movement_constraint_min.y, movement_constraint_max.y)) {
        moving_to_second = !moving_to_second;
        location.x = constrain(location.x, movement_constraint_min.x, movement_constraint_max.x);
        location.y = constrain(location.y, movement_constraint_min.y, movement_constraint_max.y);
      }
      
      simulatePlayerMove();
      if (game.player.last_moved_by_block != frameCount && getPlayerCollision()) {
        onCollide();
        //game.player.location.add(new PVector(location.x-plocation.x, location.y-plocation.y));
        if (moving_to_second) {
          game.player.location.add(movement_axis);
        } else {
          game.player.location.sub(movement_axis);
        }
        
        //println(location.x-plocation.x, location.y-plocation.y);
        //game.player.location.add((new PVector(movement_axis.x, movement_axis.y)).setMag(movement_axis.mag()*1.5));
        game.player.last_moved_by_block = frameCount;
        //println(millis()+" "+getPlayerCollision());
        
        //fill(0);
        //text(location.y-plocation.y, location.x, location.y+50);
        //text(movement_axis.y, location.x, location.y+100);
        
        //println(game.player.location.y);
      }
      unsimulatePlayerMove();
      
    }
  }
  
  int[] getBlockRounding() {
    int[] rounding = {10, 10, 10, 10}; //tl, tr, br, bl
    
    int x = round(location.x/30);
    int y = round(location.y/30);
    
    if (game.blocks.get(new PVector(x, y-1)) != null && game.blocks.get(new PVector(x, y-1)) != this && game.blocks.get(new PVector(x, y-1)).visible==this.visible) { //top
      rounding[0] = 0;
      rounding[1] = 0;
    }
    if (game.blocks.get(new PVector(x, y+1)) != null && game.blocks.get(new PVector(x, y+1)) != this && game.blocks.get(new PVector(x, y+1)).visible==this.visible) { //bottom
      rounding[2] = 0;
      rounding[3] = 0;
    }
    if (game.blocks.get(new PVector(x-1, y)) != null && game.blocks.get(new PVector(x-1, y)) != this && game.blocks.get(new PVector(x-1, y)).visible==this.visible) { //left
      rounding[0] = 0;
      rounding[3] = 0;
    }
    if (game.blocks.get(new PVector(x+1, y)) != null && game.blocks.get(new PVector(x+1, y)) != this && game.blocks.get(new PVector(x+1, y)).visible==this.visible) { //right
      rounding[1] = 0;
      rounding[2] = 0;
    }
    return rounding;
  }
  
  void simulatePlayerMove() {
    float total_x_offset = 0;
    if (game.player.a_pressed) {
      total_x_offset -= game.player.x_speed;
    }
    if (game.player.d_pressed) {
      total_x_offset += game.player.x_speed;
    }
    game.player.velocity.x += total_x_offset;
    game.player.location.y += game.player.velocity.y*game.dt+game.player.g;
  }
  
  void unsimulatePlayerMove() {
    float total_x_offset = 0;
    if (game.player.a_pressed) {
      total_x_offset -= game.player.x_speed;
    }
    if (game.player.d_pressed) {
      total_x_offset += game.player.x_speed;
    }
    game.player.velocity.x -= total_x_offset;
    game.player.location.y -= game.player.velocity.y*game.dt+game.player.g;
  }
  
  boolean getPlayerCollision() {
    return CollisionDetector.rectRect(game.player.location.x, game.player.location.y, game.player.rectSize.x, game.player.rectSize.y, location.x, location.y, block_size.x, block_size.y);
  }
  
  boolean getSimulatedPlayerCollision() {
    simulatePlayerMove();
    boolean collision = getPlayerCollision();
    unsimulatePlayerMove();
    return collision;
  }
  
  void setVertices() {
    vertices = new PVector[4];
    vertices[0] = new PVector(-block_size.x/2, -block_size.y/2);
    vertices[1] = new PVector(block_size.x/2, -block_size.y/2);
    vertices[2] = new PVector(block_size.x/2, block_size.y/2);
    vertices[3] = new PVector(-block_size.x/2, block_size.y/2);
  }
  
  boolean getPlayerSideCollision() {
    return getOutsideCollision(game.player.location.x, game.player.location.y, game.player.rectSize.x, game.player.rectSize.y);
  }
  
  boolean getOutsideCollision(float x, float y, float w, float h) {
    if (vertices == null) {
      setVertices();
    }
    
    PVector[] new_vertices = new PVector[vertices.length];
    for (int i = 0; i < new_vertices.length; i++) {
      new_vertices[i] = new PVector(location.x+vertices[i].x, location.y+vertices[i].y);
    }
    return CollisionDetector.polyRect(new_vertices, x, y, w-2, h-2);
  }
  
  Block clone() {
    try {
      return (Block) super.clone();
    } catch (CloneNotSupportedException e) {
      println("Block clone not supported");
      exit();
    }
    return null;
  }
}

class NormalBlock extends Block {
  NormalBlock(float x, float y, String _img_name) {
    location = new PVector(x, y);
    img_name = _img_name;
    block_size = new PVector(gt(img_name).width, gt(img_name).height);
    is_img = true;
    collidable = true;
    //super.init();
  }
  
  NormalBlock(float x, float y, color _color, float w, float h) {
    location = new PVector(x, y);
    rectColor = _color;
    block_size = new PVector(w, h);
    is_img = false;
    collidable = true;
    //super.init();
  }
  
  NormalBlock(float x, float y) {
    location = new PVector(x, y);
    rectColor = color(255, 127, 39);
    block_size = new PVector(30, 30);
    is_img = false;
    collidable = true;
    //super.init();
  }
  
  void update() {
    super.update();
    display();
  }
}

class DisappearingBlock extends Block {
  DisappearingBlock(float x, float y) {
    location = new PVector(x, y);
    rectColor = color(255, 127, 39);
    block_size = new PVector(30, 30);
    is_img = false;
    collidable = false;
    visible = true;
    //super.init();
  }
  
  void update() {
    super.update();
    //if (true && getSimulatedPlayerCollision()) {
    //  visible = false;
    //}
    display();
  }
  
  void onCollide() {
    visible = false;
  }
}

class AppearingBlock extends Block {
  //boolean hidden;
  AppearingBlock(float x, float y) {
    location = new PVector(x, y);
    rectColor = color(255, 127, 39);
    block_size = new PVector(30, 30);
    is_img = false;
    collidable = true;
    visible = false;
    //super.init();
  }
  
  void update() {
    super.update();
    display();
  }
  
  void onCollide() {
    visible = true;
    PVector[] coords = {new PVector(location.x/30-1, location.y/30), new PVector(location.x/30+1, location.y/30), new PVector(location.x/30, location.y/30-1), new PVector(location.x/30, location.y/30+1)};
    for (PVector coord : coords) {
      Block block = game.blocks.get(coord.copy());
      if (block != null) {
        block.block_rounding = null;
      }
    }
  }
}

class SpikeBlock extends Block {
  String direction;
  SpikeBlock(float x, float y, String _direction) {
    location = new PVector(x, y);
    direction = _direction;
    rectColor = color(255, 127, 39);
    block_size = new PVector(30, 30);
    is_img = false;
    collidable = false;
  }
  
  void update() {
    super.update();
    display();
  }
  
  void onCollide() {
    visible = true;
    game.player.die("spike");
  }
  
  void display() {
    if (!visible) {
      return;
    }
    fill(0);
    if (direction.equals("up") || direction.equals("w")) {
      triangle(location.x-block_size.x/2, location.y+block_size.y/2, location.x+block_size.x/2, location.y+block_size.y/2, location.x, location.y-block_size.y/2);
    } else if (direction.equals("down") || direction.equals("s")) {
      triangle(location.x-block_size.x/2, location.y-block_size.y/2, location.x+block_size.x/2, location.y-block_size.y/2, location.x, location.y+block_size.y/2);
    } else if (direction.equals("left") || direction.equals("a")) {
      triangle(location.x+block_size.x/2, location.y-block_size.y/2, location.x+block_size.x/2, location.y+block_size.y/2, location.x-block_size.x/2, location.y);
    } else if (direction.equals("right") || direction.equals("d")) {
      triangle(location.x-block_size.x/2, location.y-block_size.y/2, location.x-block_size.x/2, location.y+block_size.y/2, location.x+block_size.x/2, location.y);
    }
    
    //if (vertices == null) {
    //  setVertices();
    //}
    //fill(255, 0, 0);
    //pushMatrix();
    //translate(location.x, location.y);
    //stroke(255, 0, 0);
    //line(vertices[0].x, vertices[0].y, vertices[1].x, vertices[1].y);
    //line(vertices[0].x, vertices[0].y, vertices[2].x, vertices[2].y);
    //line(vertices[2].x, vertices[2].y, vertices[1].x, vertices[1].y);
    //popMatrix();
    //noStroke();
  }
  
  void setVertices() {
    vertices = new PVector[3];
    
    block_size.mult(0.8);
    if (direction.equals("up") || direction.equals("w")) {
      vertices[0] = new PVector(-block_size.x/2, block_size.y/2);
      vertices[1] = new PVector(block_size.x/2, block_size.y/2);
      vertices[2] = new PVector(0, -block_size.y/2);
    } else if (direction.equals("down") || direction.equals("s")) {
      vertices[0] = new PVector(-block_size.x/2, -block_size.y/2);
      vertices[1] = new PVector(+block_size.x/2, -block_size.y/2);
      vertices[2] = new PVector(0, block_size.y/2);
    } else if (direction.equals("left") || direction.equals("a")) {
      vertices[0] = new PVector(block_size.x/2, -block_size.y/2);
      vertices[1] = new PVector(block_size.x/2, block_size.y/2);
      vertices[2] = new PVector(-block_size.x/2, 0);
    } else if (direction.equals("right") || direction.equals("d")) {
      vertices[0] = new PVector(-block_size.x/2, -block_size.y/2);
      vertices[1] = new PVector(-block_size.x/2, +block_size.y/2);
      vertices[2] = new PVector(block_size.x/2, 0);
    }
    block_size.div(0.8);
  }
}

class SpringBlock extends Block {
  String direction;
  float rotation;
  SpringBlock(float x, float y, String _direction, float _power) {
    location = new PVector(x, y);
    rectColor = color(255, 127, 39);
    block_size = new PVector(30, 30);
    is_img = true;
    collidable = true;
    direction = _direction;
    properties.setFloat("power", _power);
    img_name = "spring.png";
    //super.init();
    rotation = get_rotation(direction);
  }
  
  void update() {
    super.update();
    boolean collided = false;
    if (direction.equals("up") || direction.equals("t")) {
      if (CollisionDetector.rectRect(game.player.location.x, game.player.location.y, game.player.rectSize.x-2, game.player.rectSize.y-2, location.x, location.y-block_size.y/2-5, block_size.x, 10)) {
        game.player.velocity.y = -properties.getFloat("power");
        collided = true;
      }
    }
    if (direction.equals("left") || direction.equals("r")) {
      if (CollisionDetector.rectRect(game.player.location.x, game.player.location.y, game.player.rectSize.x-2, game.player.rectSize.y-2, location.x-block_size.x/2-6, location.y, 12, block_size.y)) {
        game.player.velocity.x = -properties.getFloat("power")*2;
        collided = true;
      }
    }
    if (direction.equals("right") || direction.equals("y")) {
      if (CollisionDetector.rectRect(game.player.location.x, game.player.location.y, game.player.rectSize.x-2, game.player.rectSize.y-2, location.x+block_size.x/2+6, location.y, 12, block_size.y)) {
        game.player.velocity.x = properties.getFloat("power")*2;
        collided = true;
      }
    }
    if (collided) {
      visible = true;
      game.sound_engine.gs("spring2").play();
    }
    display();
  }
  
  boolean onEnemyCollide(Enemy enemy) {
    //enemy.is_dead = true;
    if (direction.equals("up") || direction.equals("t")) {
      enemy.velocity.y = -properties.getFloat("power");
      game.sound_engine.gs("spring2").play();
    } else if (direction.equals("left") || direction.equals("r")) {
      enemy.velocity.x = -properties.getFloat("power")*2;
      game.sound_engine.gs("spring2").play();
    } else if (direction.equals("right") || direction.equals("y")) {
      enemy.velocity.x = properties.getFloat("power")*2;
      game.sound_engine.gs("spring2").play();
    }
    return false;
  }
  
  void display() {
    if (!visible) {
      return;
    }
    pushMatrix();
    translate(location.x, location.y);
    rotate(rotation);
    image(gt(img_name), 0, 0);
    popMatrix();
  }
  
  float get_rotation(String rotation) {
    if (rotation.equals("up") || rotation.equals("t")) {
      return PI;
    } else if (rotation.equals("left") || rotation.equals("r")) {
      return PI/2;
    } else if (rotation.equals("down") || rotation.equals("y")) {
      return -PI/2;
    }
    return 0;
  }
}

class ShooterBlock extends Block {
  String direction;
  float bullet_speed, rotation;
  
  ShooterBlock(float x, float y, String _direction) {
    location = new PVector(x, y);
    rectColor = color(255, 127, 39);
    block_size = new PVector(30, 30);
    is_img = true;
    collidable = true;
    direction = _direction;
    properties.setFloat("full_cooldown", 120);
    properties.setFloat("cooldown", properties.getFloat("full_cooldown"));
    bullet_speed = 15;
    //super.init();
    rotation = get_rotation(direction);
    img_name = "shooter.png";
  }
  
  ShooterBlock(float x, float y, String _direction, float _cooldown) {
    location = new PVector(x, y);
    rectColor = color(255, 127, 39);
    block_size = new PVector(30, 30);
    is_img = true;
    collidable = true;
    direction = _direction;
    properties.setFloat("full_cooldown", _cooldown);
    properties.setFloat("cooldown", properties.getFloat("full_cooldown"));
    bullet_speed = 15;
    //super.init();
    rotation = get_rotation(direction);
    img_name = "shooter.png";
  }
  
  void update() {
    super.update();
    properties.setFloat("cooldown", properties.getFloat("cooldown")-game.dt);
    if (properties.getFloat("cooldown") <= 0) {
      properties.setFloat("cooldown", properties.getFloat("full_cooldown"));
      shoot();
    }
    display();
  }
  
  void shoot() {
    if (direction.equals("down") || direction.equals("S")) {
      game.bullets.add(new Arrow(location.x, location.y, 0, bullet_speed, direction));
    } else if (direction.equals("up") || direction.equals("W")) {
      game.bullets.add(new Arrow(location.x, location.y, 0, -bullet_speed, direction));
    } else if (direction.equals("left") || direction.equals("A")) {
      game.bullets.add(new Arrow(location.x, location.y, -bullet_speed, 0, direction));
    } else {
      game.bullets.add(new Arrow(location.x, location.y, bullet_speed, 0, direction));
    }
  }
  
  void display() {
    if (!visible) {
      return;
    }
    pushMatrix();
    translate(location.x, location.y);
    rotate(rotation);
    image(gt(img_name), 0, 0);
    popMatrix();
  }
  
  void onCollide() {
    visible = true;
  }
  
  float get_rotation(String rotation) {
    if (rotation.equals("up") || rotation.equals("W")) {
      return PI;
    } else if (rotation.equals("left") || rotation.equals("A")) {
      return PI/2;
    } else if (rotation.equals("down") || rotation.equals("S")) {
      return 0;
    }
    return PI*3/2;
  }
}

class PistonBlock extends Block {
  boolean inverted;
  PistonBlock(float x, float y, float _interval, boolean _inverted) {
    location = new PVector(x, y);
    rectColor = color(255, 127, 39);
    block_size = new PVector(30, 30);
    is_img = false;
    collidable = !_inverted;
    visible = true;
    properties.setFloat("full_cooldown", _interval);
    properties.setFloat("cooldown", _interval);
    inverted = _inverted;
    must_update = true;
  }
  
  void update() {
    super.update();
    
    if (properties.getFloat("cooldown") <= 0) {
      collidable = !collidable;
      visible = !visible;
      properties.setFloat("cooldown", properties.getFloat("full_cooldown"));
    } else {
      properties.setFloat("cooldown", properties.getFloat("cooldown")-game.dt);
    }
    
    display();
  }
}

class DodgingBlock extends Block {
  boolean recurrent, proximity, moving;
  PVector original_location;
  DodgingBlock(float x, float y, boolean _recurrent, boolean _proximity) {
    location = new PVector(x, y);
    original_location = new PVector(x, y);
    recurrent = _recurrent;
    proximity = _proximity;
    rectColor = color(255, 127, 39);
    block_size = new PVector(30, 30);
    is_img = false;
    moving = false;
    collidable = false;
    //super.init();
  }
  
  void update() {
    super.update();
    simulatePlayerMove();
    if ((proximity && dist(location.x, location.y, game.player.location.x, game.player.location.y) <= 150 && game.player.velocity.y < 0 && abs(game.player.location.x-location.x)>abs(game.player.location.x+game.player.velocity.x-location.x)) || CollisionDetector.rectRect(game.player.location.x, game.player.location.y, game.player.rectSize.x-2, game.player.rectSize.y-2, location.x, location.y, block_size.x-2, block_size.y-2)) {
      location.y = game.player.location.y-game.player.rectSize.y/2-block_size.y/2;
      if (proximity && location.y > original_location.y) {
        location.y = original_location.y;
      }
      if (!recurrent && proximity && original_location.y > location.y) {
        original_location.y = location.y;
      }
      moving = true;
    } else if (recurrent && moving) {
      if (location.y >= original_location.y) {
        location.y = original_location.y;
        moving = false;
      } else {
        location.y = game.player.location.y-game.player.rectSize.y/2-block_size.y/2;
      }
    } else if (!recurrent && moving) {
      float offset = 30;
      location.y = constrain(location.y-game.player.g, ceil((location.y-game.player.g)/30)*30-offset, ceil((location.y-game.player.g)/30)*30-offset);
      if (ceil(location.y/30)*30 == int(location.y/30)*30) {
        moving = false;
      }
      if (proximity) {
        original_location.y = round(original_location.y/30)*30;
        location.y = original_location.y;
      }
    }
    unsimulatePlayerMove();
    display();
    //fill(0);
    //text(round(location.y), location.x, location.y);
  }
}

class TeleportBlock extends Block {
  PVector destination;
  boolean keep_x;
  TeleportBlock(float x, float y, float dest_x, float dest_y) {
    location = new PVector(x, y);
    destination = new PVector(dest_x, dest_y);
    rectColor = color(255, 127, 39);
    block_size = new PVector(30, 30);
    is_img = false;
    collidable = false;
    visible = false;
    keep_x = false;
  }
  
  TeleportBlock(float x, float y, float dest_x, float dest_y, boolean _keep_x) {
    location = new PVector(x, y);
    destination = new PVector(dest_x, dest_y);
    rectColor = color(255, 127, 39);
    block_size = new PVector(30, 30);
    is_img = false;
    collidable = false;
    visible = false;
    keep_x = _keep_x;
  }
  
  //void update() {
  //  super.update();
  //  display();
  //}
  
  void onCollide() {
    if (!keep_x) {
      game.player.location.x = destination.x;
    }
    game.player.location.y = destination.y+17;//+block_size.y/2-game.player.rectSize.y/2;
  }
  
  //void display() {
  //  noFill();
  //  stroke(0, 0, 255);
  //  rect(location.x, location.y, block_size.x-3, block_size.y-3);
  //  noStroke();
  //}
}

class SignTriggerBlock extends Block {
  float block_id, sign_id;
  boolean triggered;
  //SignTriggerBlock(float x, float y, float _block_id, float _sign_id) {
  SignTriggerBlock(float x, float y, float _sign_id) {
    location = new PVector(x, y);
    rectColor = color(255, 127, 39);
    block_size = new PVector(30, 30);
    is_img = false;
    collidable = false;
    visible = false;
    triggered = false;
    block_id = random(-1000000000, 1000000000);
    sign_id = _sign_id;
  }
  
  SignTriggerBlock(float x, float y, float x_size, float y_size, float _sign_id) {
    location = new PVector(x, y);
    rectColor = color(255, 127, 39);
    block_size = new PVector(x_size, y_size);
    is_img = false;
    collidable = false;
    visible = false;
    triggered = false;
    block_id = random(-1000000000, 1000000000);
    sign_id = _sign_id;
  }
  
  void update() {
    super.update();
    display();
  }
  
  void onCollide() {
    triggered = true;
    for (Entity entity : game.entities) {
      if (sign_id == entity.id && block_id != entity.properties.getFloat("block_id")) {
        entity.properties.setInt("msg_index", entity.properties.getInt("msg_index")+1);
        entity.properties.setFloat("block_id", block_id);
      }
    }
  }
  
  void display() {
    //noFill();
    //stroke(0, 0, 255);
    //rect(location.x, location.y, block_size.x-3, block_size.y-3);
    //noStroke();
  }
}

class EnemySpawnBlock extends Block {
  boolean triggered;
  ArrayList<Enemy> enemy_list;
  
  EnemySpawnBlock(float x, float y, ArrayList<Enemy> _enemy_list) {
    location = new PVector(x, y);
    enemy_list = _enemy_list;
    rectColor = color(255, 127, 39);
    block_size = new PVector(30, 30);
    is_img = false;
    collidable = false;
    visible = false;
    triggered = false;
  }
  
  EnemySpawnBlock(float x, float y, int x_size, int y_size, ArrayList<Enemy> _enemy_list) {
    location = new PVector(x, y);
    enemy_list = _enemy_list;
    rectColor = color(255, 127, 39);
    block_size = new PVector(x_size, y_size);
    is_img = false;
    collidable = false;
    visible = false;
    triggered = false;
  }
  
  void onCollide() {
    if (!triggered) {
      triggered = true;
      for (int i = 0; i < enemy_list.size(); i++) {
        game.enemies.add(enemy_list.get(i));
      }
    }
  }
}

class TextBlock extends Block {
  boolean triggered;
  Text text;
  TextBlock(float x, float y, Text _text) {
    location = new PVector(x, y);
    text = _text;
    rectColor = color(255, 127, 39);
    block_size = new PVector(30, 30);
    is_img = false;
    collidable = false;
    visible = false;
    triggered = false;
  }
  
  void onCollide() {
    if (!triggered) {
      triggered = true;
      game.ui_text.add(text);
    }
  }
}

class GameCrashBlock extends Block {
  boolean triggered;
  GameCrashBlock(float x, float y) {
    location = new PVector(x, y);
    rectColor = color(255, 127, 39);
    block_size = new PVector(30, 30);
    is_img = false;
    collidable = false;
    visible = false;
    triggered = false;
  }
  
  void onCollide() {
    if (!triggered) {
      triggered = true;
      game.state = "crash";
    }
  }
}

class BarrelBlock extends Block {
  boolean triggered;
  BarrelBlock(float x, float y) {
    location = new PVector(x, y+1);
    img_name = "barrel.png";
    block_size = new PVector(30, 30);
    is_img = true;
    collidable = false;
    visible = true;
    triggered = false;
  }
  
  void update() {
    super.update();
    display();
  }
  
  void onCollide() {
    if (!triggered) {
      visible = false;
      game.player.has_fuel = true;
      triggered = true;
    }
  }
}

class CactusBlock extends Block {
  //float transition_counter;
  CactusBlock(float x, float y) {
    location = new PVector(x, y+1);
    img_name = "cactus.png";
    block_size = new PVector(30, 30);
    is_img = true;
    collidable = false;
    visible = true;
    //transition_counter = 0;
  }
  
  void update() {
    super.update();
    if (properties.getFloat("cactus_transition") == constrain(properties.getFloat("cactus_transition"), 1, 255)) {
      //transition_counter += 3*game.dt;
      properties.setFloat("cactus_transition", properties.getFloat("cactus_transition")+3*game.dt);
    }
    display();
  }
  
  void display() {
    tint(255, properties.getFloat("cactus_transition"));
    image(gt("cactus.png"), location.x, location.y);
    tint(255, 255-properties.getFloat("cactus_transition"));
    image(gt("barrel.png"), location.x, location.y);
    noTint();
  }
}

class PlaceBarrelBlock extends Block {
  PVector put_location;
  PlaceBarrelBlock(float x, float y, float w, float h, float bx, float by) {
    location = new PVector(x, y);
    rectColor = color(255, 127, 39);
    block_size = new PVector(w, h);
    put_location = new PVector(bx, by);
    is_img = false;
    collidable = false;
    visible = false;
  }
  
  void update() {
    super.update();
    display();
  }
  
  void onCollide() {
    if (game.player.has_fuel) {
      int x = (int)put_location.x;
      int y = (int)put_location.y;
      game.fbcs.add(new FillBlockCommand(x, y, x, y, new CactusBlock(0, 0)));
      game.player.has_fuel = false;
    }
  }
}

class EnemyCommandBlock extends Block {
  boolean triggered;
  JSONObject command;
  EnemyCommandBlock(float x, float y, JSONObject _command) {
    location = new PVector(x, y);
    rectColor = color(255, 127, 39);
    block_size = new PVector(30, 30);
    is_img = false;
    collidable = false;
    visible = false;
    triggered = false;
    command = _command;
  }
  
  void onCollide() {
    if (!triggered) {
      JSONObject data = command.getJSONObject("data");
      if (command.getString("type").equals("set_visible")) {
        for (int i = 0; i < game.enemies.size(); i++) {
          if (game.enemies.get(i).id == data.getFloat("id")) {
            game.enemies.get(i).visible = data.getBoolean("visible");
            continue;
          }
        }
      }
      triggered = true;
    }
  }
}

class PolyCollisionBlock extends Block {
  boolean triggered;
  JSONObject command;
  int ticks;
  PolyCollisionBlock(float x, float y) {
    location = new PVector(x, y);
    rectColor = color(255, 127, 39);
    block_size = new PVector(30, 30);
    is_img = false;
    collidable = true;
    visible = true;
    triggered = false;
    ticks = 0;
  }
  
  void update() {
    super.update();
    display();
    
    visible = true;
    if (ticks < 30) {
      ticks++;
    } else {
      visible = false;
    }
  }
  
  void display() {
    visible = true;
    super.display();
    visible = false;
  }
}

class MSOD extends Block {
  MSOD(float x, float y) {
    location = new PVector(x, y);
    rectColor = color(255, 127, 39);
    block_size = new PVector(30, 30);
    is_img = false;
    collidable = true;
    visible = true;
  }
  
  void onCollide() {
    if (collidable) {
      collidable = false;
      visible = false;
      for (int i = 0; i < random(5, 10); i++) {
        game.particles.add(new Particle(location.x, location.y, new PVector(random(-3, 3), random(-15, -6)), random(7, 12), random(360), random(-5, 5), 1, "triangle", 70, color(random(100)), 1, 1, true, true, 0.7));
      }
    }
  }
}

class BlockFillBlock extends Block {
  boolean triggered, local_visible;
  Block block_type;
  int x_fill_start, y_fill_start, x_fill_end, y_fill_end;
  BlockFillBlock(float x, float y, int _x_fill_start, int _y_fill_start, int _x_fill_end, int _y_fill_end, Block _block_type) {
    location = new PVector(x, y);
    block_size = new PVector(30, 30);
    is_img = true;
    collidable = false;
    visible = false;
    local_visible = false;
    triggered = false;
    
    x_fill_start = _x_fill_start;
    y_fill_start = _y_fill_start;
    x_fill_end = _x_fill_end;
    y_fill_end = _y_fill_end;
    block_type = _block_type;
  }
  
  BlockFillBlock(float x, float y, int _x_fill_start, int _y_fill_start, int _x_fill_end, int _y_fill_end, int x_size, int y_size, Block _block_type) {
    location = new PVector(x, y);
    block_size = new PVector(x_size, y_size);
    is_img = true;
    collidable = false;
    visible = false;
    local_visible = false;
    triggered = false;
    
    x_fill_start = _x_fill_start;
    y_fill_start = _y_fill_start;
    x_fill_end = _x_fill_end;
    y_fill_end = _y_fill_end;
    block_type = _block_type;
  }
  
  void update() {
    if (visible) {
      visible = false;
      local_visible = true;
    }
    super.update();
    display();
  }
  
  void onCollide() {
    if (!triggered) {
      triggered = true;
      game.fbcs.add(new FillBlockCommand(x_fill_start, y_fill_start, x_fill_end, y_fill_end, block_type));
    }
  }
  
  void display() {
    //noFill();
    //stroke(255, 0, 0);
    //rect(location.x, location.y, block_size.x, block_size.y);
    if (local_visible) {
      if (triggered) {
        image(gt("lever_on.png"), location.x, location.y);
      } else {
        image(gt("lever_off.png"), location.x, location.y);
      }
    }
  }
}

class FillBlockCommand {
  int x_fill_start, y_fill_start, x_fill_end, y_fill_end;
  Block block_type;
  
  FillBlockCommand(int _x_fill_start, int _y_fill_start, int _x_fill_end, int _y_fill_end, Block _block_type) {
    x_fill_start = _x_fill_start;
    y_fill_start = _y_fill_start;
    x_fill_end = _x_fill_end;
    y_fill_end = _y_fill_end;
    block_type = _block_type;
  }
  
  void run() {
    for (int i = x_fill_start; i <= x_fill_end; i++) {
      for (int j = y_fill_start; j <= y_fill_end; j++) {
        //Class<?> c = Class.forName("Block");
        if (block_type != null) {
          game.blocks.put(new PVector(i, j), block_type.clone());
          game.blocks.get(new PVector(i, j)).location = new PVector(i*30, j*30);
        } else {
          game.blocks.remove(new PVector(i, j));
        }
      }
    }
  }
}