class Bullet {
  PVector location, velocity, rectSize;
  String img_name;
  boolean is_dead;
  float fuse, rotation, img_scale;
  
  void update() {}
  
  void display() {
    pushMatrix();
    translate(location.x, location.y);
    scale(img_scale);
    rotate(rotation);
    image(gt(img_name), 0, 0);
    popMatrix();
  }
  
  boolean getPlayerCollision() {
    return CollisionDetector.rectRect(game.player.location.x, game.player.location.y, game.player.rectSize.x, game.player.rectSize.y, location.x, location.y, rectSize.x, rectSize.y);
  }
  
  boolean getBlockCollision() {
    for (HashMap.Entry<PVector,Block> entry : game.blocks.entrySet()) {
      Block block = entry.getValue();
      if (abs(location.x-block.location.x) <= 150 && abs(location.y-block.location.y) <= 150 && block.getOutsideCollision(location.x, location.y, rectSize.x, rectSize.y)) {
        if (block.collidable) {
          return true;
        }
      }
    }
    return false;
  }
  
  float get_rotation(String rotation) {
    if (rotation.equals("up") || rotation.equals("W")) {
      return PI;
    } else if (rotation.equals("left") || rotation.equals("A")) {
      return PI/2;
    } else if (rotation.equals("down") || rotation.equals("S")) {
      return 0;
    }
    return PI*3/2;
  }
}

class PlayerBullet extends Bullet {
  //float img_scale;
  PlayerBullet(float x, float y, float xv, float yv, float _rotation, String _img_name, float _img_scale) {
    location = new PVector(x, y);
    velocity = new PVector(xv, yv);
    rotation = _rotation;
    img_name = _img_name;
    img_scale = _img_scale;
    rectSize = new PVector(gt(img_name).width, gt(img_name).height).mult(img_scale);
  }
  
  void update() {
    location.add(velocity);
    if (getBlockCollision()) {
      is_dead = true;
    }
    display();
  }
  
  //void display() {
  //  pushMatrix();
  //  translate(location.x, location.y);
  //  scale(img_scale);
  //  rotate(rotation);
  //  image(gt(img_name), 0, 0);
  //  popMatrix();
  //}
}

class LethalBullet extends Bullet {
  LethalBullet(float x, float y, float xv, float yv, String _img_name) {
    img_name = _img_name;
    location = new PVector(x, y);
    velocity = new PVector(xv, yv);
    rectSize = new PVector(gt(img_name).width, gt(img_name).height);
    img_scale = 1;
  }
  
  void update() {
    if (is_dead) {
      return;
    }
    move();
    display();
  }
  
  void move() {
    boolean init_block_collision = getBlockCollision();
    location.add(velocity);
    if (!init_block_collision && getBlockCollision()) {
      onBlockCollision();
      is_dead = true;
    }
    if (getPlayerCollision()) {
      game.player.die("bullet");
    }
    fuse -= game.dt;
    if (fuse <= 0) {
      onBlockCollision();
      is_dead = true;
    }
  }
  
  void onBlockCollision() {}
}

class Arrow extends LethalBullet {
  Arrow(float x, float y, float xv, float yv, String rotation_str) {
    super(x, y, xv, yv, "arrow.png");
    rotation = get_rotation(rotation_str);
    fuse = 1000;
  }
  
  void onBlockCollision() {
    for (int i = 0; i < random(5, 10); i++) {
      //game.particles.add(new Particle(location.x, location.y, new PVector(random(velocity.x-3, velocity.x+3), random(velocity.y-3, velocity.y+3)), random(7, 12)*img_scale, random(360), random(-5, 5), 1, "triangle", 70, color(185, 122, 87), 1, 1, false, false, 0));
      game.particles.add(new Particle(location.x, location.y, new PVector(random(constrain(velocity.x, -1, 1)-1, constrain(velocity.x, -1, 1)+1), random(constrain(velocity.y, -1, 1)-1, constrain(velocity.y, -1, 1)+1)), random(3, 7)*img_scale, random(360), random(-5, 5), 1, "triangle", 70, color(185, 122, 87), 1, 1, false, false, 0));
    }
  }
}

class FireBall extends LethalBullet {
  boolean explosive, affected_by_gravity;
  color theColor;
  int g_increase, g_min, g_max;
  float size;
  FireBall(float x, float y, float xv, float yv, float _size, boolean _explosive, boolean _affected_by_gravity) {
    super(x, y, xv, yv, "void.png");
    size = _size;
    rectSize = new PVector(size, size);
    explosive = _explosive;
    affected_by_gravity = _affected_by_gravity;
    fuse = 1000;
    g_increase = 5;
    g_min = 50;
    g_max = 150;
    theColor = color(255, (g_min+g_max)/2, 0);
  }
  
  void update() {
    super.update();
  }
  
  void move() {
    if (affected_by_gravity) {
      velocity.y += game.player.g/20;
    }
    super.move();
  }
  
  void display() {
    theColor = color(red(theColor), green(theColor)+g_increase, blue(theColor));
    if (green(theColor) != constrain(green(theColor), g_min+1, g_max)) {
      g_increase *= -1;
      color(red(theColor), constrain(green(theColor), g_min, g_max), blue(theColor));
    }
    
    fill(theColor);
    ellipse(location.x, location.y, size, size);
    spawn_particles();
  }
  
  void spawn_particles() {
    game.particles.add(new Particle(location.x+random(-rectSize.x/2, rectSize.x/2), location.y+random(-rectSize.y/2, rectSize.y/2), (new PVector(random(-1, 1), random(-1, 1))), 10, 0, 0, 0, "circle", 70, color(255, random(g_min, g_max), 0, 120), 0.5, 1, false, false, 0));
  }
  
  void onBlockCollision() {
    if (!explosive) {
      return;
    }
    for (int i = 0; i < 10; i++) {
      game.bullets.add(new FireBall(location.x-velocity.x, location.y-velocity.y, velocity.mag()*cos(radians(i*36)), velocity.mag()*sin(radians(i*36)), size*0.33, false, true));
    }
  }
}

class GuidedMissle extends LethalBullet {
  float size, speed, guidance, particles_per_tick;
  boolean aggroed;
  GuidedMissle(float x, float y, float _guidance, float _fuse) {
    super(x, y, 0, -6, "missle.png");
    guidance = _guidance;
    //size = 10;
    speed = 6;
    rectSize = new PVector(gt(img_name).width, gt(img_name).height);
    fuse = _fuse;
    img_scale = 2;
    particles_per_tick = 3;
    aggroed = false;
  }
  
  void move() {
    if (!aggroed && dist(location.x, location.y, game.player.location.x, game.player.location.y) < dist(0, 0, width/2, height/2)) {
      aggroed = true;
    }
    if (aggroed) {
      PVector destination = new PVector(game.player.location.x-location.x, game.player.location.y-location.y).setMag(speed*guidance);
      //velocity.lerp(destination, guidance);
      velocity.add(destination);
      rotation = velocity.heading();
      
      location.add(velocity);
      if (getPlayerCollision()) {
        game.player.die("bullet");
        explode();
        //is_dead = true;
      }
      fuse -= game.dt;
      if (fuse <= 0) {
        explode();
        is_dead = true;
      }
    }
  }
  
  void display() {
    super.display();
    spawn_particles();
  }
  
  void explode() {
    for (int i = 0; i < 10; i++) {
      game.bullets.add(new FireBall(location.x-velocity.x, location.y-velocity.y, 0.1*velocity.x+5*cos(radians(i*36)), 0.1*velocity.y+5*sin(radians(i*36)), size*0.33, false, true));
    }
    game.sound_engine.gs("explosion").play();
  }
  
  void spawn_particles() {
    for (int i = 0; i < particles_per_tick; i++) {
      PVector tail_location = (new PVector(-rectSize.x, 5)).rotate(rotation).add(location);
      PVector particle_speed = (new PVector(-speed/3, 0)).rotate(rotation+radians(random(-30, 30)));
      game.particles.add(new Particle(tail_location.x, tail_location.y, particle_speed, 10, 0, 0, 0, "circle", 70, color(0, 120), 0.5, 1, false, false, 0));
    }
  }
}

class FallingDeath extends LethalBullet {
  //float img_scale, rotation_speed;
  float rotation_speed;
  FallingDeath(float x, float y, String img_name, float _img_scale) {
    super(x, y, 0, 5, img_name);
    img_scale = _img_scale;
    rectSize.mult(img_scale);
    rotation = random(TWO_PI);
    fuse = 1000;
    rotation_speed = random(5, 10)*((round(random(1))*2)-1);
  }
  
  void update() {
    velocity.y += game.player.g*0.5;
    rotation += radians(rotation_speed);
    super.update();
  }
  
  //void display() {
  //  pushMatrix();
  //  translate(location.x, location.y);
  //  rotate(rotation);
  //  scale(img_scale);
  //  image(gt(img_name), 0, 0);
  //  popMatrix();
  //}
  
  void onBlockCollision() {
    for (int i = 0; i < random(15, 25); i++) {
      game.particles.add(new Particle(location.x, location.y, new PVector(random(-3, 3), random(-10, -3)), 10, 0, 0, 0, "circle", 70, color(255, 0, 0, 120), 0.5, 1, true, true, 0.8));
    }
  }
}

class Anvil extends LethalBullet {
  //float img_scale;
  boolean active;
  Anvil(float x, float y, float _img_scale) {
    super(x, y, 0, 0, "anvil.png");
    img_scale = _img_scale;
    rectSize.mult(img_scale);
    fuse = 1000000000;
    active = false;
  }
  
  void update() {
    if (!active && (location.x+rectSize.x/2 >= game.player.location.x-game.player.rectSize.x/2 && location.x-rectSize.x/2 <= game.player.location.x+game.player.rectSize.x/2)) {
      active = true;
    }
    if (active) {
      velocity.y += game.player.g;
    }
    super.update();
  }
  
  //void display() {
  //  pushMatrix();
  //  translate(location.x, location.y);
  //  scale(img_scale);
  //  image(gt(img_name), 0, 0);
  //  popMatrix();
  //}
  
  void onBlockCollision() {
    for (int i = 0; i < random(5, 10); i++) {
      game.particles.add(new Particle(location.x, location.y, new PVector(random(-3, 3), random(-15, -6)), random(7, 12)*img_scale, random(360), random(-5, 5), 1, "triangle", 70, color(0), 1, 1, true, true, 0.7));
    }
    game.sound_engine.gs("anvil").play();
  }
}