class Enemy {
  PVector location, velocity, rectSize;
  String img_name;
  boolean is_dead, visible;
  float id, g_cfnt;
  
  Enemy() {
    is_dead = false;
    visible = true;
    id = random(-1000000000, 1000000000);
    g_cfnt = 0.5;
  }
  
  void update() {}
  
  void display() {}
}

class BasicEnemy extends Enemy {
  float speed, rotation, bounce, img_scale;
  String collider;
  boolean aggroed;
  int direction_facing;
  BasicEnemy(float x, float y, float _speed, String _img_name) {
    location = new PVector(x, y);
    velocity = new PVector(0, 0);
    speed = _speed;
    img_name = _img_name;
    rectSize = new PVector(gt(img_name).width, gt(img_name).height);
    aggroed = false;
    img_scale = 1;
    direction_facing = 1;
  }
  
  void update() {
    move();
    display();
  }
  
  void move() {
    if (!aggroed && dist(location.x, location.y, game.player.location.x, game.player.location.y) < width/2+height/2) {
      aggroed = true;
    }
    if (!aggroed) {
      return;
    }
    location.x += (velocity.x+speed)*game.dt;
    if (collidesWithBlocks()) {
      location.x -= (velocity.x+speed)*game.dt;
      velocity.x = 0;
      speed *= -1;
    }
    velocity.x *= 0.99;
    
    
    velocity.y += game.player.g*g_cfnt;
    location.y += velocity.y*game.dt;
    if (collidesWithBlocks()) {
      location.y -= velocity.y*game.dt;
      velocity.y *= -bounce;
      if (abs(velocity.y) <= 2) {
        velocity.y = 0;
      } else {
        game.sound_engine.gs("boing").play();
      }
    }
    velocity.y = constrain(velocity.y, -20, 20);
    
    if (collidesWithPlayer() && !game.player.is_dead) {
      game.player.die("enemy");
    }
    
    if (speed < 0) {
      direction_facing = 1;
    } else if (speed > 0) {
      direction_facing = -1;
    }
  }
  
  void display() {
    if (!visible) {
      return;
    }
    pushMatrix();
    translate(location.x, location.y);
    scale(direction_facing*img_scale, img_scale);
    rotate(radians(rotation));
    image(gt(img_name), 0, 0);
    popMatrix();
  }
  
  boolean collidesWithBlocks() {
    for (HashMap.Entry<PVector,Block> entry : game.blocks.entrySet()) {
      Block block = entry.getValue();
      if (abs(location.x-block.location.x) <= 150 && abs(location.y-block.location.y) <= 150 && block.getOutsideCollision(location.x, location.y, rectSize.x*img_scale, rectSize.y*img_scale)) {
        //if (block.collidable) {
        boolean true_collision = block.onEnemyCollide(this);
        if (true_collision && block.visible) {
          return true;
        }
        //block.onEnemyCollide(this); //make the polyblock visible again!
      }
    }
    return false;
  }
  
  boolean collidesWithPlayer() {
    if (collider.equals("circle")) {
      if (CollisionDetector.circleRect(location.x, location.y, rectSize.x*img_scale/2, game.player.location.x, game.player.location.y, game.player.rectSize.x, game.player.rectSize.y)) {
        onPlayerCollide();
        return true;
      }
    } else if (collider.equals("rect")) {
      if (CollisionDetector.rectRect(location.x, location.y, rectSize.x*img_scale, rectSize.y*img_scale, game.player.location.x, game.player.location.y, game.player.rectSize.x, game.player.rectSize.y)) {
        onPlayerCollide();
        return true;
      }
    }
    return false;
  }
  
  void onPlayerCollide() {
    speed = 0;
  }
}

class HappyFace extends BasicEnemy {
  float rotation_scale;
  HappyFace(float x, float y) {
    super(x, y, -2, "happy_face.png");
    collider = "circle";
    bounce = 0;
    rotation_scale = 1;
  }
  
  HappyFace(float x, float y, float _bounce) {
    super(x, y, -2, "happy_face.png");
    collider = "circle";
    bounce = _bounce;
    rotation_scale = 1;
  }
  
  HappyFace(float x, float y, float _bounce, float _rotation_scale, float _img_scale) {
    super(x, y, -2, "happy_face.png");
    collider = "circle";
    bounce = _bounce;
    rotation_scale = _rotation_scale;
    img_scale = _img_scale;
  }
  
  void update() {
    super.update();
    rotation -= rotation_scale;
  }
}

class TrollFarmer extends BasicEnemy {
  boolean collided;
  float final_x;
  TrollFarmer(float x, float y) {
    super(x, y, -3.5, "troll_farmer.png");
    collider = "rect";
    bounce = 0;
    collided = false;
  }
  
  void update() {
    rotation = random(-10, 10);
    location.y -= random(5);
    super.update();
    
    if (collided) {
      if (location.x < final_x) {
        speed = 0;
      }
    }
  }
  
  void onPlayerCollide() {
    if (!collided) {
      game.sound_engine.gs("smack").play();
      collided = true;
      while (!game.player.collidesWithBlocks()) {
        game.player.location.y++;
      }
      game.player.display_sideways = true;
      final_x = random(game.player.location.x-30, game.player.location.x+30);
      Particle[] blood_particles = {new Particle(0, 0, new PVector(0, 0), 10, 0, 0, 0, "circle", 100, color(236, 30, 61, 100), 0, 1, true, false, 0)};
      game.particle_systems.add(new ParticleSystem(game.player.location.x, game.player.location.y, blood_particles, 1));
    }
  }
}

class FlyingCat extends BasicEnemy {
  float cooldown, max_cooldown, img_scale;
  FlyingCat (float x, float y) {
    super(x, y, 1, "cat.png");
    velocity = new PVector(-speed, 0);
    collider = "none";
    bounce = 0;
    img_scale = 3;
    cooldown = 0;
    max_cooldown = 60;
  }
  
  void update() {
    move();
    shoot();
    display();
  }
  
  void move() {
    if (!aggroed && dist(location.x, location.y, game.player.location.x, game.player.location.y) < width/2+height/2) {
      aggroed = true;
    }
    if (!aggroed) {
      return;
    }
    velocity.rotate(radians(random(-15, 15)));
    location.x += velocity.x;
    if (collidesWithBlocks()) {
      location.x -= velocity.x;
      velocity.x *= -1;
    }
    location.y += velocity.y;
    if (collidesWithBlocks()) {
      location.y -= velocity.y;
      velocity.y *= -1;
    }
  }
  
  void shoot() {
    if (!aggroed && dist(location.x, location.y, game.player.location.x, game.player.location.y) < width/2+height/2) {
      aggroed = true;
    }
    if (!aggroed) {
      return;
    }
    if (cooldown <= 0) {
      cooldown = max_cooldown;
      PVector bullet_velocity = new PVector(game.player.location.x-location.x, game.player.location.y-location.y).setMag(3);
      game.bullets.add(new FireBall(location.x, location.y, bullet_velocity.x, bullet_velocity.y, 20, true, false));
    } else {
      cooldown -= game.dt;
    }
  }
}

class Frog extends BasicEnemy {
  float move_cooldown, max_move_cooldown;
  Frog(float x, float y) {
    super(x, y, 0, "frog.png");
    velocity = new PVector(0, 0);
    collider = "rect";
    bounce = 0;
    move_cooldown = 0;
    max_move_cooldown = 60;
  }
  
  void move() {
    if (velocity.y == 0) {
      location.y += game.player.g/2;
      if (collidesWithBlocks()) {
        if (move_cooldown <= 0) {
          //velocity = new PVector(-3, -8);
          velocity = new PVector(-4, -12);
        } else {
          move_cooldown -= game.dt;
        }
      }
      location.y -= game.player.g/2;
    }
    
    
    location.x += velocity.x;
    if (collidesWithBlocks()) {
      location.x -= velocity.x;
    }
    
    velocity.y += game.player.g/2;
    location.y += velocity.y;
    if (collidesWithBlocks()) {
      location.y -= velocity.y;
      if (velocity.y > game.player.g/2) {
        move_cooldown = max_move_cooldown;
        velocity.x = 0;
      }
      velocity.y = 0;
    }
  }
}

class FireBreathingFrog extends Frog {
  float cooldown, max_cooldown;
  boolean triggered;
  FireBreathingFrog(float x, float y) {
    super(x, y);
    cooldown = 0;
    max_cooldown = 6;
    triggered = false;
  }
  
  void update() {
    super.update();
    shoot();
  }
  
  void shoot() {
    if (!triggered && dist(location.x, location.y, game.player.location.x, game.player.location.y) <= width/8) {
      triggered = true;
    }
    
    if (cooldown <= 0) {
      if (triggered) {
        cooldown = max_cooldown;
        PVector bullet_velocity = new PVector(game.player.location.x-location.x, game.player.location.y-location.y).setMag(6);
        game.bullets.add(new FireBall(location.x, location.y, bullet_velocity.x, bullet_velocity.y, 5, false, false));
      }
    } else {
      cooldown -= game.dt;
    }
  }
}