class Entity {
  PVector location, rectSize;
  boolean is_dead;
  float id;
  JSONObject properties;
  
  Entity() {
    properties = new JSONObject();
    is_dead = false;
    id = random(-1000000000, 1000000000);
  }
  
  void update() {}
  
  void display() {}
}

class Sign extends Entity {
  String[] msgs;
  String img_name;
  float img_scale, text_size, display_img_scale;
  Sign(float x, float y, String[] _msg, float _img_scale, float _text_size) {
    location = new PVector(x, y);
    rectSize = new PVector(gt("sign.png").width, gt("sign.png").height);
    msgs = _msg;
    img_scale = _img_scale;
    properties.setInt("msg_index", 0);
    properties.setFloat("block_id", 0);
    img_name = "";
    text_size = _text_size;
  }
  
  Sign(float x, float y, String[] _msg, float _text_size) {
    location = new PVector(x, y);
    rectSize = new PVector(gt("sign.png").width, gt("sign.png").height);
    msgs = _msg;
    img_scale = 1;
    properties.setInt("msg_index", 0);
    properties.setFloat("block_id", 0);
    img_name = "";
    text_size = _text_size;
  }
  
  Sign(float x, float y, String _img_name, float _display_img_scale, float _img_scale) {
    location = new PVector(x, y);
    rectSize = new PVector(gt("sign.png").width, gt("sign.png").height);
    msgs = null;
    img_scale = _img_scale;
    properties.setInt("msg_index", 0);
    properties.setFloat("block_id", 0);
    img_name = _img_name;
    display_img_scale = _display_img_scale;
  }
  
  void update() {
    display();
  }
  
  void display() {
    pushMatrix();
    translate(location.x, location.y);
    scale(img_scale);
    image(gt("sign.png"), 0, 0);
    popMatrix();
    if (img_name != "") {
      pushMatrix();
      translate(location.x, location.y-img_scale*7);
      scale(display_img_scale);
      image(gt(img_name), 0, 0);
      popMatrix();
    }
    
    if (msgs != null) {
      pushStyle();
      textFont(game.sign_font);
      textSize(text_size);
      textAlign(CENTER, CENTER);
      fill(0);
      text(msgs[constrain(properties.getInt("msg_index"), 0, msgs.length-1)], location.x, location.y-rectSize.y/2);
      popStyle();
    }
  }
}

class RainCloud extends Entity {
  PVector velocity;
  float img_scale, cooldown, max_cooldown;
  boolean can_follow, following;
  RainCloud(float x, float y, float xv, float yv, float _img_scale, float interval, boolean _can_follow) {
    location = new PVector(x, y);
    velocity = new PVector(xv, yv);
    img_scale = _img_scale;
    rectSize = new PVector(gt("cloud.png").width*img_scale, gt("cloud.png").height*img_scale);
    cooldown = 0;
    max_cooldown = interval;
    can_follow = _can_follow;
    following = false;
  }
  
  void update() {
    move();
    spawn();
    display();
  }
  
  void move() {
    if (can_follow && !following && location.x >= game.player.location.x - 20 && location.x <= game.player.location.x + 20) {
      following = true;
    }
    
    if (!following) {
      location.add(velocity);
    } else {
      location.x = game.player.location.x;
    }
  }
  
  void spawn() {
    if (cooldown <= 0) {
      String[] bullet_imgs = {"cat.png", "dog.png"};
      game.bullets.add(new FallingDeath(location.x+random(-rectSize.x/2, rectSize.x/2), location.y+random(-rectSize.y/2, rectSize.y/2), bullet_imgs[round(random(0, 1))], random(0.5, 1.5)*img_scale));
      cooldown = max_cooldown;
    } else {
      cooldown -= game.dt;
    }
  }
  
  void display() {
    pushMatrix();
    translate(location.x, location.y);
    scale(img_scale);
    image(gt("cloud.png"), 0, 0);
    popMatrix();
  }
}

class CheckPoint extends Entity {
  String img_name;
  CheckPoint(float x, float y) {
    location = new PVector(x, y);
    img_name = "checkpoint.png";
    rectSize = new PVector(gt(img_name).width, gt(img_name).height);
  }
  
  void update() {
    checkCollision();
    display();
  }
  
  void checkCollision() {
    if (checkPlayerCollision()) {
      game.player.checkpoint = new PVector(location.x+rectSize.x/2, location.y-rectSize.y/2);
    }
  }
  
  void display() {
    image(gt(img_name), location.x+rectSize.x/2, location.y-rectSize.y/2);
  }
  
  boolean checkPlayerCollision() {
    return CollisionDetector.rectRect(location.x+rectSize.x/2, location.y-rectSize.y/2, rectSize.x, rectSize.y, game.player.location.x, game.player.location.y, game.player.rectSize.x, game.player.rectSize.y);
  }
}

class RunningCheckPoint extends CheckPoint {
  float total_dist, cur_dist;
  RunningCheckPoint(float x, float y, float dist) {
    super(x, y);
    cur_dist = 0;
    total_dist = dist;
  }
  
  void update() {
    move();
    super.update();
  }
  
  void move() {
    if (cur_dist < total_dist && checkPreemptivePlayerCollision()) {
      float offset = constrain(getPlayerXMove(), -abs(total_dist-cur_dist), abs(total_dist-cur_dist));
      location.x += offset;
      cur_dist += offset;
    }
  }
  
  boolean checkPreemptivePlayerCollision() {
    boolean collision = false;
    simulatePlayerMove();
    collision = checkPlayerCollision();
    unsimulatePlayerMove();
    return collision;
  }
  
  void simulatePlayerMove() {
    float total_x_offset = 0;
    if (game.player.a_pressed) {
      total_x_offset -= game.player.x_speed;
    }
    if (game.player.d_pressed) {
      total_x_offset += game.player.x_speed;
    }
    game.player.location.x += total_x_offset;
    game.player.location.y += game.player.velocity.y*game.dt+game.player.g;
  }
  
  void unsimulatePlayerMove() {
    float total_x_offset = 0;
    if (game.player.a_pressed) {
      total_x_offset -= game.player.x_speed;
    }
    if (game.player.d_pressed) {
      total_x_offset += game.player.x_speed;
    }
    game.player.location.x -= total_x_offset;
    game.player.location.y -= game.player.velocity.y*game.dt+game.player.g;
  }
  
  float getPlayerXMove() {
    float p1 = game.player.location.x;
    simulatePlayerMove();
    float p2 = game.player.location.x;
    unsimulatePlayerMove();
    return p2-p1;
  }
}

class EndPoint extends Entity {
  String img_name;
  EndPoint(float x, float y) {
    location = new PVector(x, y);
    img_name = "endpoint.png";
    rectSize = new PVector(gt(img_name).width, gt(img_name).height);
  }
  
  void update() {
    checkCollision();
    display();
  }
  
  void checkCollision() {
    if (getCollision()) {
      //game.player.checkpoint = new PVector(location.x+rectSize.x/2, location.y-rectSize.y/2);
      game.state = "choose_level";
      if (game.level_name.equals("3")) {
        game.state = "credits";
        game.credits = new Credits(true);
      }
      String[] lb = loadStrings("levels/lb.txt");
      String[] nlb = new String[lb.length+1];
      for (int i = 0; i < lb.length; i++) {
        nlb[i] = lb[i];
      }
      nlb[nlb.length-1] = game.level_name;
      saveStrings("data/levels/lb.txt", nlb);
      game.level_name = "";
      //printArray(nlb);
    }
  }
  
  void display() {
    image(gt(img_name), location.x+rectSize.x/2, location.y-rectSize.y/2);
  }
  
  boolean getCollision() {
    return CollisionDetector.rectRect(location.x+rectSize.x/2, location.y-rectSize.y/2, rectSize.x, rectSize.y, game.player.location.x, game.player.location.y, game.player.rectSize.x, game.player.rectSize.y);
  }
}

class InvisEndpoint extends EndPoint {
  InvisEndpoint(float x, float y) {
    super(x, y);
    rectSize = new PVector(100, 100);
  }
  
  void update() {
    if (getCollision()) {
      game.state = "fade_and_win";
      return;
    }
  }
  
  void display() {}
}

class BoatMan extends Entity {
  Animation troll;
  BoatMan(float x, float y) {
    location = new PVector(x, y);
    rectSize = new PVector(0, 0);
    
    ArrayList<String> animation_phases = new ArrayList<String>();
    boolean out_of_imgs = false;
    for (int i = 1; !out_of_imgs; i++) {
      if (gt("player_"+i+".png") != null) {
        animation_phases.add("player_"+i+".png");
      } else {
        out_of_imgs = true;
      }
    }
    troll = new Animation(animation_phases, 100);
  }
  
  void update() {
    display();
  }
  
  void display() {
    pushMatrix();
    translate(location.x, location.y-15);
    scale(-1, 1);
    image(troll.ci(), 0, 0);
    popMatrix();
    
    pushMatrix();
    translate(location.x+3, location.y-46);
    scale(0.1, 0.1);
    image(gt("tophat.png"), 0, 0);
    popMatrix();
    
    image(gt("boat.png"), location.x+200, location.y-50);
  }
}

class Redav extends Entity {
  Animation animation;
  float tick_millis, last_millis;
  Redav(float x, float y) {
    location = new PVector(x, y);
    rectSize = new PVector(0, 0);
    tick_millis = 600;
    last_millis = millis();
    
    
    ArrayList<String> animation_phases = new ArrayList<String>();
    for (int i = 1; i < 9; i++) {
      animation_phases.add("redav_"+min(i, 10-i)+".png");
    }
    animation = new Animation(animation_phases, tick_millis/3);
  }
  
  void update() {
    display();
  }
  
  void display() {
    pushMatrix();
    float offset = -(millis() - tick_millis - last_millis)*0.1;
    translate(location.x, location.y-140-min(tick_millis*0.1-offset, offset));
    image(animation.ci(), 0, 0);
    popMatrix();
    
    while (millis() - tick_millis > last_millis) {
      last_millis += tick_millis;
    }
  }
}

class Portal extends Entity {
  PVector[] locations;
  Portal(float x1, float y1, float x2, float y2) {
    locations = new PVector[2];
    locations[0] = new PVector(x1, y1);
    locations[1] = new PVector(x2, y2);
    rectSize = new PVector(40, 60);
  }
  
  void update() {
    teleport();
    display();
  }
  
  void teleport() {
    if (CollisionDetector.rectRect(locations[0].x+rectSize.x/2, locations[0].y-rectSize.y/2, rectSize.x, rectSize.y, game.player.location.x, game.player.location.y, game.player.rectSize.x, game.player.rectSize.y)) {
      game.player.location = locations[1].copy();
    }
  }
  
  void display() {
    for (PVector cur_location : locations) {
      pushMatrix();
      translate(cur_location.x, cur_location.y);
      
      pushStyle();
      colorMode(HSB, 100);
      
      stroke((millis()/10)%100, 100, 100);
      noFill();
      ellipse(0, 0, rectSize.x, rectSize.y);
      
      noStroke();
      for (int i = 10; i <= 200; i += 10) {
        fill((i+millis()/10)%100, 100, 100, 30);
        float counter = i*0.05+(millis())*0.01;
        triangle(0, 0, rectSize.x/2*cos(counter), rectSize.y/2*sin(counter), rectSize.x/2*sin(counter), rectSize.y/2*cos(counter));
      }
      
      popStyle();
      
      popMatrix();
    }
  }
}