class Game {
  float dt, player_dead_counter, max_player_dead_counter, loading_counter, max_loading_counter, crash_counter, introduction_counter, introduction_initial_millis, fade_counter;
  String state, pstate, player_name, introduction_state, game_over_text, helpful_hint, level_name;
  int fail_counter;
  Intro intro;
  Menu menu;
  Credits credits;
  TrollCredits troll_credits;
  SoundEngine sound_engine;
  Player player;
  HashMap<PVector,Block> blocks;
  ArrayList<Bullet> bullets;
  ArrayList<Enemy> enemies;
  ArrayList<Entity> entities;
  ArrayList<Particle> particles;
  ArrayList<ParticleSystem> particle_systems;
  ArrayList<FillBlockCommand> fbcs;
  ArrayList<Text> ui_text;
  PFont sign_font, credits_title_font, ror_font;
  boolean debug, first_hint_used;
  ArrayList<LolText> lol_text;
  ArrayList<PVector> level_coords;
  ArrayList<String> level_titles;
  Animation cyclist;
  
  Game() {
    texture_manager = new TextureManager();
    dt = 1;
    state = "intro"; //intro //choose_level
    player = new Player(0, 0);
    player_dead_counter = 0;
    player_name = "";
    introduction_counter = 0;
    introduction_state = "fade";
    introduction_initial_millis = 0;
    loading_counter = 0;
    max_loading_counter = 0;
    fade_counter = 0;
    fail_counter = 0;
    
    blocks = new HashMap<PVector,Block>();
    bullets = new ArrayList<Bullet>();
    enemies = new ArrayList<Enemy>();
    entities = new ArrayList<Entity>();
    particles = new ArrayList<Particle>();
    particle_systems = new ArrayList<ParticleSystem>();
    fbcs = new ArrayList<FillBlockCommand>();
    ui_text = new ArrayList<Text>();
    
    level_name = "";
    level_coords = new ArrayList<PVector>();
    level_coords.add(new PVector(width*0.25, height*0.3));
    level_coords.add(new PVector(width*0.5, height*0.3));
    level_coords.add(new PVector(width*0.75, height*0.3));
    level_coords.add(new PVector(width*0.25, height*0.7));
    level_coords.add(new PVector(width*0.5, height*0.7));
    level_coords.add(new PVector(width*0.75, height*0.7));
    level_titles = new ArrayList<String>();
    level_titles.add("The Beginning\nof the End");
    level_titles.add("The End of\nthe Beginning");
    level_titles.add("The Chasm");
    level_titles.add("The Kingdom\nof Lrutnylbr");
    level_titles.add("The Evil Monkey\nStatue of Discontent");
    level_titles.add("The Genie");
    
    lol_text = new ArrayList<LolText>();
    first_hint_used = false;
    
    imageMode(CENTER);
    rectMode(CENTER);
    noStroke();
    textSize(20);
    textAlign(CENTER, CENTER);
    
    
    sign_font = loadFont("fonts/sign_font.vlw");
    credits_title_font = loadFont("fonts/credits_title_font.vlw");
    ror_font = loadFont("fonts/ror.vlw");
    debug = false;
    
    sound_engine = new SoundEngine(get_papplet());
    intro = new Intro(sound_engine);
    //intro.stop_sound();
    menu = new Menu(sound_engine);
    menu.pause();
    
    //credits = new Credits();
    //troll_credits = new TrollCredits();
    
    ArrayList<String> animation_phases = new ArrayList<String>();
    boolean out_of_imgs = false;
    for (int i = 1; !out_of_imgs; i++) {
      if (gt("body_"+i+".png") != null) {
        animation_phases.add("body_"+i+".png");
      } else {
        out_of_imgs = true;
      }
    }
    cyclist = new Animation(animation_phases, 100);
  }
  
  void update() {
    dt = 60.0/frameRate;
    sound_engine.update();
    if (state.equals("intro")) {
      intro.update();
      if (intro.finished) {
        state = "menu";
        menu.unpause();
      }
    } else if (state.equals("menu")) {
      menu.update();
    } else if (state.equals("get_player_name")) {
      getPlayerName();
    } else if (state.equals("choose_level")) {
      background(153, 217, 234);
      for (int i = 0; i < level_coords.size(); i++) {
        boolean selected = false;
        strokeWeight(5);
        stroke(255);
        if (CollisionDetector.pointRect(mouseX, mouseY, level_coords.get(i).x, level_coords.get(i).y, 200, 200)) {
          stroke(255, 0, 0);
          selected = true;
        }
        image(gt((i+1)+".png"), level_coords.get(i).x, level_coords.get(i).y);
        fill(255, 0);
        rect(level_coords.get(i).x, level_coords.get(i).y, 200, 200);
        
        fill(0);
        if (selected) {
          fill(255, 0, 0);
        }
        textFont(credits_title_font);
        textSize(20);
        textAlign(CENTER, TOP);
        
        text(level_titles.get(i), level_coords.get(i).x, level_coords.get(i).y+110);
        noStroke();
      }

      if (!level_name.equals("")) {
        boolean avaliable = false;
        String[] lb = loadStrings("data/levels/lb.txt");
        for (String i : lb) {
          if (int(i)+1 == int(level_name)) {
            avaliable = true;
            break;
          }
        }
        
        //if (avaliable || level_name.equals("1")) {
        //  loadLevel(level_name, false);
        //  if (level_name.equals("1")) {
        //    state = "introduction";
        //  } else {
        //    state = "play";
        //  }
        //}
        
        if (avaliable || level_name.equals("1")) {
          state = "loading "+level_name;
        }
      }
    } else if (state.split(" ")[0].equals("loading")) {
      if (loading_counter == 0) {
        loadLevel(state.split(" ")[1], false);
        max_loading_counter = random(240, 480);
        
        String[] helpful_hint_list = 
        {
        "To win you have to beat the game",
        "Never flip off the mafia",
        "Never shave with a lawnmower",
        "Failure is the answer, and never the question",
        "Try not to break tha game this time",
        "Think this is gonna be fun? Think again",
        "Oh you are gluten free? So you go against the grain? Pasta la vista baby",
        "If you're playing this you're probably masochistic",
        "Sell at most one of your kidneys",
        "Your body has the correct number of holes in it. Don't make any more.",
        "Don't tie yourself to an airplane propeller",
        "Give me all your money",
        "Under no circumstances should you ever reproduce",
        "Don't iron clothes while wearing them"
        };
        int helpful_hint_index = (int)random(helpful_hint_list.length);
        if (!first_hint_used) {
          first_hint_used = true;
          helpful_hint_index = 0;
        }
        helpful_hint = helpful_hint_list[helpful_hint_index];
      }
      
      background(0);
      pushMatrix();
      translate(width/2, height/2-100);
      //scale(1.5, 1.5);
      image(cyclist.ci(), 0, -20);
      image(gt("head.png"), 0, -35);
      rotate(millis()*0.005);
      image(gt("mc.png"), 0, 0);
      popMatrix();
      
      textSize(40);
      fill(255, 200, 0, map(sin(degrees((millis()-introduction_initial_millis)/15000.0)), -1, 1, 0, 150));
      text("Loading...", width/2, height*0.6);
      
      textSize(20);
      fill(255, 200, 0, 150);
      text("Helpful hint: "+helpful_hint, width/2, height*0.8);
      
      loading_counter += game.dt;
      if (loading_counter >= max_loading_counter) {
        loading_counter = 0;
        if (state.split(" ")[1].equals("1")) {
          state = "introduction";
        } else {
          state = "play";
        }
      }
    } else if (state.equals("introduction")) {
      playIntroduction();
    } else if (state.equals("play")) {
      pushMatrix();
      //rotate(radians(90)); //90_1_-1
      translate(width/2-player.location.x, height/2-player.location.y);
      background(153, 217, 234);
      updateBlocks();
      updateEntities();
      player.update();
      updateOther();
      popMatrix();
      display_ui();
    } else if (state.equals("player_dead")) {
      if (player_dead_counter == 0) {
        lol_text = new ArrayList<LolText>();
        String[] game_over_text_list = 
        {
        player_name+".exe was not compatible with this version of World",
        player_name+" decided to go to a better place",
        "Why did " + player_name + " want to die? We will never know",
        player_name + " thought he was Chuck Norris",
        player_name + " died from unknown causes",
        "Only " + player_name + " could figure out how to die like that",
        player_name + " transformed into a muffin",
        player_name + " became one with the earth",
        "The world will be much better without " + player_name,
        player_name.toUpperCase() + " FAILED",
        player_name + " did a weird",
        player_name + "'s soul escaped",
        player_name + " was against net neutrality",
        player_name + " hates life",
        player_name + " took too many arrows to the knee",
        player_name + " rests in pieces",
        "One does not simply beat the game",
        player_name + " should have read that book on physics",
        player_name + " didn\'t make enough horcruxes",
        player_name + " met a giant jellyfish",
        player_name + " left a small crater",
        player_name + " was brutally murdered by " + player_name,
        player_name + " brought a nuke to a knife fight",
        player_name + " divided by their own intelligence",
        player_name + " played a rage game",
        "ur bad kid",
        player_name + " felt a tad bit suicidal",
        player_name + " thought there were only 2 genders",
        player_name + " was crucified by SJWs",
        player_name + " said 'theiyr're' to a grammar nazi",
        player_name + " got what they deserved",
        "LOL"
        };
        
        game_over_text = game_over_text_list[(int)random(game_over_text_list.length)];
      }
      
      if (game_over_text.equals("LOL")) {
        for (int i = 0; i < 5; i++) {
          lol_text.add(new LolText(random(width), random(height)));
        }
      }
      
      player_dead_counter += dt;
      
      pushMatrix();
      translate(width/2-player.location.x, height/2-player.location.y);
      background(153, 217, 234);
      displayBlocks();
      displayEntities();
      player.display();
      displayOther();
      popMatrix();
      display_ui();
      
      textFont(sign_font);
      if (!game_over_text.equals("LOL")) {
        textAlign(CENTER, TOP);
        textSize(40);
        fill(0);
        text(game_over_text, width/2, height*4/5);
        textAlign(CENTER, CENTER);
      } else {
        textAlign(CENTER, CENTER);
        textSize(30);
        fill(random(255), random(255), random(255));
        for (LolText lol_text_entry : lol_text) {
          lol_text_entry.display();
        }
      }
      
      if (player_dead_counter >= max_player_dead_counter) {
        loadLevel(level_name, true);
        PVector checkpoint = new PVector(player.checkpoint.x, player.checkpoint.y);
        player = new Player(checkpoint.x, checkpoint.y);
        player_dead_counter = 0;
        state = "play";
      }
    } else if (state.equals("fade_and_win")) {
      pushMatrix();
      translate(width/2-player.location.x, height/2-player.location.y);
      background(153, 217, 234);
      displayBlocks();
      displayEntities();
      player.display();
      displayOther();
      popMatrix();
      display_ui();
      
      fill(0, max(0, fade_counter-100));
      rect(width/2, height/2, width, height);
      fade_counter += 2*dt;
      if (fade_counter >= 355) {
        fade_counter = 0;
        String[] lb = loadStrings("levels/lb.txt");
        String[] nlb = new String[lb.length+1];
        for (int i = 0; i < lb.length; i++) {
          nlb[i] = lb[i];
        }
        nlb[nlb.length-1] = level_name;
        saveStrings("data/levels/lb.txt", nlb);
        level_name = "";
        state = "credits";
        credits = new Credits(true);
      }
    } else if (state.equals("crash")) {
      if (crash_counter > 100 && crash_counter < 120) {
        fill(255, 8);
        rect(width/2, height/2, width, height);
        surface.setTitle(get_surface_name()+" (Not Responding)");
      }
      crash_counter += game.dt;
      if (crash_counter >= 360) {
        crash_counter = 0;
        state = "play";
        surface.setTitle(get_surface_name());
      }
    } else if (state.equals("credits")) {
      credits.update();
    } else if (state.equals("troll_credits")) {
      troll_credits.update();
    } else if (state.equals("exit")) {
      exit();
    }
  }
  
  void loadLevel(String name, boolean keep_player_location) {
    blocks = new HashMap<PVector,Block>();
    bullets = new ArrayList<Bullet>();
    enemies = new ArrayList<Enemy>();
    entities = new ArrayList<Entity>();
    particles = new ArrayList<Particle>();
    particle_systems = new ArrayList<ParticleSystem>();
    fbcs = new ArrayList<FillBlockCommand>();
    ui_text = new ArrayList<Text>();
    
    //String[] text = {"g1m1", "g2m1", "g1m2"};
    //String[] names = {"guy 1", "guy 2"};
    //String[] face_img_names = {"trollface.png", "trollface_tophat.png"};
    //float[] face_img_scales = {0.5, 0.5};
    //int[] order = {0, 1, 0};
    
    //ui_text.add(new Text(text, names, face_img_names, face_img_scales, order));
    
    //for (int i = 0; i <= 100; i++) {
    //  blocks.put(new PVector(i, 0), new NormalBlock(i*30, 0));
    //}
    
    Level level_1 = (new LevelRenderer()).loadLevel(name, keep_player_location);
    blocks = level_1.blocks;
    bullets = level_1.bullets;
    enemies = level_1.enemies;
    entities = level_1.entities;
    
    
    
    //blocks.put(new PVector(5, -3), new NormalBlock(150, -90));
    //blocks.get(new PVector(5, -3)).setMoving(150, -90, 150, -290, 1);
    //blocks.put(new PVector(5, -2), new SpikeBlock(150, -60, "down"));
    ////blocks.get(new PVector(5, -2)).visible = false;
    //blocks.get(new PVector(5, -2)).setMoving(150, -60, 150, -260, 1);
    //blocks.put(new PVector(4, -3), new SpikeBlock(120, -90, "left"));
    ////blocks.get(new PVector(4, -3)).visible = false;
    //blocks.get(new PVector(4, -3)).setMoving(120, -90, 120, -290, 1);
    //blocks.put(new PVector(6, -3), new SpikeBlock(180, -90, "right"));
    ////blocks.get(new PVector(6, -3)).visible = false;
    //blocks.get(new PVector(6, -3)).setMoving(180, -90, 180, -290, 1);
    //blocks.put(new PVector(5, -4), new SpikeBlock(150, -120, "up"));
    ////blocks.get(new PVector(5, -4)).visible = false;
    //blocks.get(new PVector(5, -4)).setMoving(150, -120, 150, -320, 1);
    
    //blocks.put(new PVector(10, -2), new DisappearingBlock(300, -60));
    //blocks.put(new PVector(13, -2), new AppearingBlock(390, -60));
    
    //blocks.put(new PVector(15, 2), new ShooterBlock(450, -60, "up"));
    
    //blocks.put(new PVector(20, 3), new DodgingBlock(600, -90, true, true));
    //blocks.put(new PVector(30, 3), new DodgingBlock(900, -90, true, false));
    //blocks.put(new PVector(40, 3), new DodgingBlock(1200, -90, false, true));
    //blocks.put(new PVector(50, 3), new DodgingBlock(1500, -90, false, false));
    
    //blocks.put(new PVector(60, 3), new NormalBlock(1800, -90));
    //blocks.get(new PVector(60, 3)).setMoving(1800, -90, 1800, -300, 2);
    
    //blocks.put(new PVector(70, 3), new NormalBlock(2100, -90));
    //blocks.get(new PVector(70, 3)).setMoving(2100, -90, 2400, -90, 2);
    
    //blocks.put(new PVector(80, 3), new NormalBlock(2400, -90));
    //blocks.get(new PVector(80, 3)).setMoving(2400, -90, 2700, -300, 2);
    
    //bullets.add(new GuidedMissle(400, -200, 0.04));
    
    //for (int i = 0; i < 20; i++) {
    //  bullets.add(new Anvil(400+i*30, -400, 1));
    //}
    
    //for (int i = 1; i < 10; i++) {
    //  blocks.put(new PVector(34, -i), new AppearingBlock(1020, -i*30));
    //}
    //blocks.put(new PVector(25, 1), new BlockFillBlock(750, -60, 34, -10, 34, -1, new NormalBlock(0, 0, color(255, 127, 39), 30, 30)));
    
    //for (int i = 0; i < 5; i++) {
    //  enemies.add(new TrollFarmer(400+5*i, -50));
    //}
    //enemies.add(new FlyingCat(400, -200));
    //bullets.add(new FireBall(400, -200, -2, 2, 20, true, false));
    //bullets.add(new GuidedMissle(400, -200, 0.02));
    
    //bullets.add(new FallingDeath(400, -500, "cat.png", 3));
    //entities.add(new RainCloud(400, -300, 0, 0, 1, 20, false));
    //entities.add(new RainCloud(400, -300, -1, 0, 2, 20));
    
    //String[] sign_msgs = {"THIS WAY\n----->\n", "THIS WAY\n----->\n(to death)"};
    //entities.add(new Sign(400, -165, sign_msgs, 5));
    //blocks.put(new PVector(15, -1), new SignTriggerBlock(450, -30, entities.get(entities.size()-1).id));
    
    //enemies.add(new FireBreathingFrog(400, -200));
    
    
    
    //for (int i = -10; i <= 10; i++) {
    //  for (int j = 0; j >= -120; j -= 30) {
    //    blocks.put(new PVector(i, j), new NormalBlock(i*30, j*30, color(255, 127, 39), 30, 30));
    //  }
    //}
    
    //for (int j = -1; j > -120; j -= 30) {
    //  int neg_x = 1;
    //  String spike_direction = "right";
    //  int x_off = 0;
    //  if ((j+1)%60==0 || j==-91) {
    //    neg_x = -1;
    //    spike_direction = "left";
    //    if (j == -91) {
    //      x_off = -3;
    //    }
    //  }
      
    //  blocks.put(new PVector(neg_x*9+x_off, j), new TeleportBlock(neg_x*240+x_off*30, j*30, 0, 30*(j-30)-30, true));
    //  for (int i = 0; i <= 10; i++) {
    //    blocks.put(new PVector(-neg_x*9, j-i), new BlockFillBlock(-neg_x*300, (j-i)*30, -neg_x*10, j-i, -neg_x*10, j, new SpikeBlock(0, 0, spike_direction)));
    //  }
    //}
    
    //blocks.put(new PVector(-11, -121), new BlockFillBlock(-11*30, -121*30, -51, -120, -11, -120, new NormalBlock(0, 0)));
    
    //String[] sign_msgs = {"This way\n<-----"};
    //entities.add(new Sign(0, -105, sign_msgs.clone(), 3));
    //sign_msgs[0] = "No wait, it's this\nway ----->";
    //entities.add(new Sign(0, -105-900, sign_msgs.clone(), 3));
    //sign_msgs[0] = "Actually, it's this\n<----- way";
    //entities.add(new Sign(0, -105-1800, sign_msgs.clone(), 3));
    //sign_msgs[0] = "|      This      |\nV     way?     V";
    //entities.add(new Sign(0, -105-2700, sign_msgs.clone(), 3));
    //sign_msgs[0] = "middle\nfinger";
    //entities.add(new Sign(0, -105-3600, "middle_finger.png", 3));
  }
  
  void updateBlocks() {
    for (HashMap.Entry<PVector,Block> entry : blocks.entrySet()) {
      Block block = entry.getValue();
      if (block.must_update || (abs(player.location.x-block.location.x) <= width && abs(player.location.y-block.location.y) <= height)) {
        entry.getValue().update(); 
      }
    }
    
    for (int i = fbcs.size()-1; i >= 0; i--) {
      fbcs.get(i).run();
      fbcs.remove(i);
    }
  }
  
  void updateEntities() {
    for (int i = entities.size()-1; i >= 0; i--) {
      entities.get(i).update();
      if (entities.get(i).is_dead) {
        entities.remove(i);
      }
    }
  }
  
  void updateOther() {
    for (int i = bullets.size()-1; i >= 0; i--) {
      bullets.get(i).update();
      if (bullets.get(i).is_dead) {
        bullets.remove(i);
      }
    }
    
    for (int i = enemies.size()-1; i >= 0; i--) {
      enemies.get(i).update();
      if (enemies.get(i).is_dead) {
        enemies.remove(i);
      }
    }
    
    for (int i = particle_systems.size()-1; i >= 0; i--) {
      particle_systems.get(i).update();
      if (particle_systems.get(i).is_dead) {
        particle_systems.remove(i);
      }
    }
    
    for (int i = particles.size()-1; i >= 0; i--) {
      particles.get(i).update();
      if (particles.get(i).is_dead) {
        particles.remove(i);
      }
    }
  }
  
  void displayBlocks() {
    for (HashMap.Entry<PVector,Block> entry : blocks.entrySet()) {
      Block block = entry.getValue();
      if (block.must_update || (abs(player.location.x-block.location.x) <= width && abs(player.location.y-block.location.y) <= height)) {
        entry.getValue().display(); 
      }
    }
  }
  
  void displayEntities() {
    for (int i = 0; i < entities.size(); i++) {
      entities.get(i).display();
    }
  }
  
  void displayOther() {
    //for (int i = bullets.size()-1; i >= 0; i--) {
    //  bullets.get(i).update();
    //  if (bullets.get(i).is_dead) {
    //    bullets.remove(i);
    //  }
    //}
    
    for (int i = 0; i < bullets.size(); i++) {
      bullets.get(i).display();
    }
    
    for (int i = 0; i < enemies.size(); i++) {
      enemies.get(i).update();
    }
    
    for (int i = particle_systems.size()-1; i >= 0; i--) {
      particle_systems.get(i).update();
      if (particle_systems.get(i).is_dead) {
        particle_systems.remove(i);
      }
    }
    
    for (int i = particles.size()-1; i >= 0; i--) {
      particles.get(i).update();
      if (particles.get(i).is_dead) {
        particles.remove(i);
      }
    }
  }
  
  void display_ui() {
    for (int i = ui_text.size()-1; i >= 0; i--) {
      ui_text.get(i).display();
      if (ui_text.get(i).is_dead) {
        ui_text.remove(i);
      }
    }
    
    pushStyle();
    textFont(sign_font);
    fill(0);
    textSize(20);
    textAlign(RIGHT, TOP);
    text("Fails: "+fail_counter, width-10, 10);
    
    if (!debug) {
      return;
    }
    textAlign(LEFT, TOP);
    text("FPS: "+round(frameRate), 10, 10);
    text("DT: "+round(dt*10)/10.0, 120, 10);
    text("LX: "+round(player.location.x), 10, 35);
    text("LY: "+round(player.location.y), 120, 35);
    text("BX: "+round(player.location.x/30), 10, 60);
    text("BY: "+round(player.location.y/30), 120, 60);
    text("VX: "+round(player.velocity.x*100)/100, 10, 85);
    text("VY: "+round(player.velocity.y*100)/100, 120, 85);
    popStyle();
  }
  
  void getPlayerName() {
    background(map(sin(degrees(millis()/12345.0)), -1, 1, 50, 255), map(sin(degrees(millis()/23456.0)), -1, 1, 50, 255), map(sin(degrees(millis()/34567.0)), -1, 1, 50, 255));
    textFont(credits_title_font);
    textAlign(CENTER, CENTER);
    textSize(40*map(noise(millis()/1000.0), 0, 1, 1, 1.1));
    fill(255-map(sin(degrees(millis()/12345.0)), -1, 1, 50, 255), 255-map(sin(degrees(millis()/23456.0)), -1, 1, 50, 255), 255-map(sin(degrees(millis()/34567.0)), -1, 1, 50, 255));
    text("Enter you name, heathen:", width/2, height/4);
    text(player_name, width/2, height*3/4);
  }
  
  void playIntroduction() {
    if (introduction_state.equals("fade")) {
      if (introduction_counter == 0) {
        introduction_counter = 1;
      }
      fill(0, introduction_counter);
      rect(width/2, height/2, width, height);
      introduction_counter *= 1.05;
      if (introduction_counter >= 50) {
        introduction_state = "scroll";
        introduction_counter = 0;
      }
    } else if (introduction_state.equals("scroll")) {
      if (introduction_counter >= -1050) {
        background(0);
        pushMatrix();
        translate(width/2, introduction_counter);
        
        textFont(credits_title_font);
        textAlign(CENTER, CENTER);
        textSize(30);
        fill(255, 200, 0);
        text("A long time ago in the not so distant future,\nthe tale of "+player_name+" began. I, the narrator\nwith no voice acting at the time, will tell you\nthe tale of the things that will have been and\nwill have not been been "+player_name+" doing not\nthe stuff it did. The tale was sad,\n(but short), but "+player_name+" knew this not. All\nit knew was how to word properly sentences. Be\namazed by its tale of bravery and stupidity\nas we travel from his future into its past.\nOr something like that.", 0, height*3/2-50);
        
        popMatrix();
        introduction_counter -= game.dt/2;
        
        if (introduction_counter <= -700) {
          if (introduction_initial_millis == 0) {
            introduction_initial_millis = millis();
          }
          textSize(25);
          fill(255, 200, 0, map(sin(degrees((millis()-introduction_initial_millis)/15000.0)), -1, 1, 0, 150));
          text("Press any key to continue...", width/2, height*0.9);
        }
        
        //textFont(game.sign_font);
        //textAlign(LEFT, BOTTOM);
        //fill(255);
        //textSize(30);
        //text(round(introduction_counter), 20, height-20);
      } else {
        introduction_counter = 1;
        introduction_state = "fade";
        state = "play";
        introduction_initial_millis = 0;
      }
    }
  }
  
  void keyPressed() {
    if (state.equals("play") || state.equals("crash")) {
      player.keyPressed();
    }
    if (keyCode == ESC) {
      key = 0;
      if (state.equals("play")) {
        state = "menu";
        //player_name = "";
        level_name = "";
      }
    }
  }
  
  void keyReleased() {
    if (state.equals("play") || state.equals("crash")) {
      player.keyReleased();
    } else if (state.equals("intro")) {
      state = "menu";
      intro.stop_sound();
      menu.unpause();
    } else if (state.equals("get_player_name")) {
      if ("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890".indexOf(key) != -1) {
        player_name += key;
      } else if (keyCode == BACKSPACE && player_name.length() > 0) {
        player_name = player_name.substring(0, player_name.length()-1);
      } else if (keyCode == ENTER) {
        state = "choose_level";
      }
    } else if (state.equals("introduction")) {
      introduction_counter = -10000;
      introduction_state = "scroll";
    }
  }
  
  void keyTyped() {
    if (state.equals("play") || state.equals("crash")) {
      if (key == '\\') {
        debug = !debug;
      }
      if (key == ' ') {
        for (int i = 0; i < ui_text.size(); i++) {
          ui_text.get(i).mouseClicked();
        }
      }
    } else if (state.equals("troll_credits") && troll_credits != null) {
      troll_credits.keyTyped();
    }
  }
  
  void mouseClicked() {
    if (state.equals("menu")) {
      menu.mouseClicked();
    } else if (state.equals("choose_level")) {
      for (int i = 0; i < level_coords.size(); i++) {
        if (CollisionDetector.pointRect(mouseX, mouseY, level_coords.get(i).x, level_coords.get(i).y, 200, 200)) {
          level_name = str(i+1);
          return;
        }
      }
    } else if (state.equals("play")) {
      for (int i = 0; i < ui_text.size(); i++) {
        ui_text.get(i).mouseClicked();
      }
    }
  }
  
  String renderGameVer(String str) {
    String[] parts = str.split("_");
    String rstr = "";
    for (int i = 1; i < parts.length; i++) {
      rstr += parts[i]+'.';
    }
    rstr = rstr.substring(0, rstr.length()-1);
    
    rstr = "Version " + rstr;
    return rstr;
  }
}