class Level {
  HashMap<PVector,Block> blocks;
  ArrayList<Bullet> bullets;
  ArrayList<Enemy> enemies;
  ArrayList<Entity> entities;
  
  Level(HashMap<PVector,Block> _blocks, ArrayList<Bullet> _bullets, ArrayList<Enemy> _enemies, ArrayList<Entity> _entities) {
    blocks = _blocks;
    bullets = _bullets;
    enemies = _enemies;
    entities = _entities;
  }
}

class LevelRenderer {
  Level loadLevel(String level_name, boolean keep_player_location) {
    HashMap<PVector,Block> blocks = new HashMap<PVector,Block>();
    ArrayList<Bullet> bullets = new ArrayList<Bullet>();
    ArrayList<Enemy> enemies = new ArrayList<Enemy>();
    ArrayList<Entity> entities = new ArrayList<Entity>();
    
    String[] raw_contents = loadStrings("levels/"+level_name+".txt");
    ArrayList<ArrayList<String>> contents = new ArrayList<ArrayList<String>>();
    for (String line : raw_contents) {
      ArrayList<String> line_contents = new ArrayList<String>();
      for (String chtr : line.split("")) {
        line_contents.add(chtr);
      }
      contents.add(line_contents);
    }
    
    boolean found = false;
    PVector origin = new PVector(-1, -1);
    for (int i = 0; i < contents.size(); i++) {
      for (int j = 0; j < contents.get(i).size(); j++) {
        if (contents.get(i).get(j).equals("X")) {
          found = true;
          origin = new PVector(j, i);
        }
        if (found) {
          break;
        }
      }
      if (found) {
        break;
      }
    }
    
    //HashMap<PVector,String> rendered_contents = new HashMap<PVector,String>();
    //for (int i = 0; i < contents.size(); i++) {
    //  for (int j = 0; j < contents.get(i).size(); j++) {
    //    rendered_contents.put(new PVector(i-origin.x, j-origin.y), contents.get(i).get(j));
    //  }
    //}
    
    for (int i = 0; i < contents.size(); i++) {
      for (int j = 0; j < contents.get(i).size(); j++) {
        if (contents.get(i).get(j).equals("N")) {
          bullets.add(new Anvil(30*(j-origin.x), 30*(i-origin.y), 1));
        } else if (contents.get(i).get(j).equals("C")) {
          entities.add(new CheckPoint(30*(j-origin.x), 30*(i-origin.y+0.5)));
          //entities.add(new RunningCheckPoint(30*(j-origin.x), 30*(i-origin.y+1), 200));
        } else if (contents.get(i).get(j).equals("E")) {
          entities.add(new EndPoint(30*(j-origin.x), 30*(i-origin.y+1)));
        } else if (contents.get(i).get(j).equals("B")) {
          entities.add(new BoatMan(30*(j-origin.x), 30*(i-origin.y)));
          entities.add(new InvisEndpoint(30*(j-origin.x)+200, 30*(i-origin.y)-50));
        } else if (contents.get(i).get(j).equals("R")) {
          entities.add(new Redav(30*(j-origin.x), 30*(i-origin.y)));
        } else if (contents.get(i).get(j).equals("g")) {
          bullets.add(new GuidedMissle(30*(j-origin.x), 30*(i-origin.y), 0.1, 300));
        } else if (contents.get(i).get(j).equals("G")) {
          bullets.add(new GuidedMissle(30*(j-origin.x), 30*(i-origin.y), 0.1, 300));
          bullets.get(bullets.size()-1).img_name = "moosle.png";
        } else if ("1234567890".indexOf(contents.get(i).get(j)) != -1 && !contents.get(i).get(j).equals("")) {
          Enemy enemy = get_enemy(contents.get(i).get(j), 30*(j-origin.x), 30*(i-origin.y));
          if (enemy != null) {
            enemies.add(enemy);
          }
        } else {
          Block block = get_block(contents.get(i).get(j), 30*(j-origin.x), 30*(i-origin.y));
          if (block != null) {
            blocks.put(new PVector(j-origin.x, i-origin.y), block);
          }
        }
      }
    }
    
    JSONArray block_commands = loadJSONArray("levels/"+level_name+"_data.json");
    for (int i = 0; i < block_commands.size(); i++) {
      JSONObject command = block_commands.getJSONObject(i);
      String operand = command.getString("operand");
      String type = command.getString("type");
      JSONObject data = command.getJSONObject("data");
      if (operand.equals("block")) {
        if (type.equals("set_moving")) {
          blocks.get(new PVector(data.getInt("bx"), data.getInt("by"))).setMoving(data.getFloat("x1"), data.getFloat("y1"), data.getFloat("x2"), data.getFloat("y2"), data.getFloat("speed"));
        } else if (type.equals("set_visible")) {
          blocks.get(new PVector(data.getInt("bx"), data.getInt("by"))).visible = data.getBoolean("visible");
        } else if (type.equals("set_visible_chunk")) {
          for (int bx = data.getInt("bx1"); bx <= data.getInt("bx2"); bx++) {
            for (int by = data.getInt("by1"); by <= data.getInt("by2"); by++) {
              blocks.get(new PVector(bx, by)).visible = data.getBoolean("visible");
            }
          }
        } else if (type.equals("set_collidable")) {
          blocks.get(new PVector(data.getInt("bx"), data.getInt("by"))).collidable = data.getBoolean("collidable");
        } else if (type.equals("set_collidable_chunk")) {
          for (int bx = data.getInt("bx1"); bx <= data.getInt("bx2"); bx++) {
            for (int by = data.getInt("by1"); by <= data.getInt("by2"); by++) {
              blocks.get(new PVector(bx, by)).collidable = data.getBoolean("collidable");
            }
          }
        } else if (type.equals("set_must_update")) {
          blocks.get(new PVector(data.getInt("bx"), data.getInt("by"))).must_update = data.getBoolean("must_update");
        } else if (type.equals("set_power")) {
          blocks.get(new PVector(data.getInt("bx"), data.getInt("by"))).properties.setFloat("power", data.getFloat("power"));
        } else if (type.equals("set_shooting_speed")) {
          blocks.get(new PVector(data.getInt("bx"), data.getInt("by"))).properties.setFloat("full_cooldown", data.getFloat("speed"));
          blocks.get(new PVector(data.getInt("bx"), data.getInt("by"))).properties.setFloat("cooldown", data.getFloat("speed"));
        } else if (type.equals("set_interval")) {
          blocks.get(new PVector(data.getInt("bx"), data.getInt("by"))).properties.setFloat("full_cooldown", data.getFloat("interval"));
          blocks.get(new PVector(data.getInt("bx"), data.getInt("by"))).properties.setFloat("cooldown", data.getFloat("interval"));
        } else if (type.equals("set_teleporter")) {
          blocks.put(new PVector(data.getInt("bx"), data.getInt("by")), new TeleportBlock(data.getInt("bx")*30, data.getInt("by")*30, data.getFloat("dx"), data.getFloat("dy"), data.getBoolean("keep_x")));
        } else if (type.equals("set_sign_trigger")) {
          if (data.isNull("x_size")) {
            blocks.put(new PVector(data.getInt("bx"), data.getInt("by")), new SignTriggerBlock(data.getInt("bx")*30, data.getInt("by")*30, data.getFloat("sign_id")));
          } else {
            blocks.put(new PVector(data.getInt("bx"), data.getInt("by")), new SignTriggerBlock(data.getInt("bx")*30, data.getInt("by")*30, data.getFloat("x_size"), data.getFloat("y_size"), data.getFloat("sign_id")));
          }
        } else if (type.equals("set_enemy_spawn")) {
          ArrayList<Enemy> enemies_to_spawn = new ArrayList<Enemy>();
          JSONArray enemy_list = data.getJSONArray("enemies");
          for (int j = 0; j < enemy_list.size(); j++) {
            JSONObject cur_enemy = enemy_list.getJSONObject(j);
            float x = cur_enemy.getFloat("x");
            if (!data.isNull("x_range")) {
              x += random(-data.getFloat("x_range")/2, data.getFloat("x_range")/2);
            }
            float y = cur_enemy.getFloat("y");
            if (!data.isNull("y_range")) {
              y += random(-data.getFloat("y_range")/2, data.getFloat("y_range")/2);
            }
            enemies_to_spawn.add(get_enemy(cur_enemy.getString("type"), x, y));
          }
          if (data.isNull("x_size")) {
            blocks.put(new PVector(data.getInt("bx"), data.getInt("by")), new EnemySpawnBlock(data.getInt("bx")*30, data.getInt("by")*30, enemies_to_spawn));
          } else {
            blocks.put(new PVector(data.getInt("bx"), data.getInt("by")), new EnemySpawnBlock(data.getInt("bx")*30, data.getInt("by")*30, data.getInt("x_size"), data.getInt("y_size"), enemies_to_spawn));
          }
        } else if (type.equals("set_text_block")) {
          if (!(keep_player_location && !data.isNull("intro") && data.getBoolean("intro"))) {
            String split_key = "\\"+data.getString("key");
            String[] text = data.getString("text").split(split_key);
            String[] names = data.getString("names").split(split_key);
            String[] face_img_names = data.getString("face_img_names").split(split_key);
            String[] face_img_scales_str = data.getString("face_img_scales").split(split_key);
            String[] order_str = data.getString("order").split(split_key);
            
            float[] face_img_scales = new float[face_img_scales_str.length];
            for (int j = 0; j < face_img_scales_str.length; j++) {
              face_img_scales[j] = float(face_img_scales_str[j]);
            }
            
            int[] order = new int[order_str.length];
            for (int j = 0; j < order_str.length; j++) {
              order[j] = int(order_str[j]);
            }
            
            blocks.put(new PVector(data.getInt("bx"), data.getInt("by")), new TextBlock(data.getInt("bx")*30, data.getInt("by")*30, new Text(text, names, face_img_names, face_img_scales, order)));
          }
        } else if (type.equals("set_block_fill")) {
          if (data.isNull("x_size")) {
            blocks.put(new PVector(data.getInt("bx"), data.getInt("by")), new BlockFillBlock(data.getInt("bx")*30, data.getInt("by")*30, data.getInt("x1"), data.getInt("y1"), data.getInt("x2"), data.getInt("y2"), get_block(data.getString("block_type"), 0, 0)));
          } else {
            blocks.put(new PVector(data.getInt("bx"), data.getInt("by")), new BlockFillBlock(data.getInt("bx")*30, data.getInt("by")*30, data.getInt("x1"), data.getInt("y1"), data.getInt("x2"), data.getInt("y2"), data.getInt("x_size"), data.getInt("y_size"), get_block(data.getString("block_type"), 0, 0)));
          }
        } else if (type.equals("set_enemy_command_block")) {
          blocks.put(new PVector(data.getInt("bx"), data.getInt("by")), new EnemyCommandBlock(data.getInt("bx")*30, data.getInt("by")*30, data.getJSONObject("command")));
        } else if (type.equals("set_place_barrel_block")) {
          blocks.put(new PVector(data.getInt("bx"), data.getInt("by")), new PlaceBarrelBlock(data.getInt("bx")*30, data.getInt("by")*30, data.getInt("bw"), data.getInt("bh"), data.getInt("pbx"), data.getInt("pby")));
        }
      } else if (operand.equals("bullet")) {
        if (type.equals("arrow")) {
          bullets.add(new Arrow(data.getFloat("x"), data.getFloat("y"), data.getFloat("xv"), data.getFloat("yv"), data.getString("rotation")));
        } else if (type.equals("fireball")) {
          bullets.add(new FireBall(data.getFloat("x"), data.getFloat("y"), data.getFloat("xv"), data.getFloat("yv"), data.getFloat("size"), data.getBoolean("explosive"), data.getBoolean("affected_by_gravity")));
        } else if (type.equals("missle")) {
          bullets.add(new GuidedMissle(data.getFloat("x"), data.getFloat("y"), data.getFloat("guidance"), data.getFloat("fuse")));
        } else if (type.equals("falling_death")) {
          bullets.add(new FallingDeath(data.getFloat("x"), data.getFloat("y"), data.getString("img_name"), data.getFloat("img_scale")));
        } else if (type.equals("anvil")) {
          bullets.add(new Anvil(data.getFloat("x"), data.getFloat("y"), data.getFloat("img_scale")));
        } else if (type.equals("set_fuse")) {
          bullets.get(bullets.size()-1).fuse = data.getFloat("fuse");
        } else if (type.equals("set_rotation")) {
          bullets.get(bullets.size()-1).rotation = radians(data.getFloat("rotation"));
        } else if (type.equals("set_img_scale")) {
          bullets.get(bullets.size()-1).img_scale = data.getFloat("img_scale");
        }
      } else if (operand.equals("enemy")) {
        boolean spawned = false;
        if (type.equals("happy_face")) {
          enemies.add(new HappyFace(data.getFloat("x"), data.getFloat("y"), data.getFloat("bounce"), data.getFloat("rotation_scale"), data.getFloat("scale")));
          spawned = true;
        } else if (type.equals("troll_farmer")) {
          enemies.add(new TrollFarmer(data.getFloat("x"), data.getFloat("y")));
          spawned = true;
        } else if (type.equals("flying_cat")) {
          enemies.add(new FlyingCat(data.getFloat("x"), data.getFloat("y")));
          spawned = true;
        } else if (type.equals("frog")) {
          enemies.add(new Frog(data.getFloat("x"), data.getFloat("y")));
          spawned = true;
        } else if (type.equals("fire_breathing_frog")) {
          enemies.add(new FireBreathingFrog(data.getFloat("x"), data.getFloat("y")));
          spawned = true;
        } else if (type.equals("set_visible")) {
          for (int j = 0; j < enemies.size(); j++) {
            if (enemies.get(j).id == data.getFloat("id")) {
              enemies.get(j).visible = data.getBoolean("visible");
              continue;
            }
          }
        }
        
        if (!data.isNull("id") && spawned) {
          enemies.get(enemies.size()-1).id = data.getFloat("id");
        }
      } else if (operand.equals("entity")) {
        boolean spawned = false;
        if (type.equals("sign")) {
          if (data.getString("sign_type").equals("text")) {
            String [] msgs = data.getString("msg").split("\\"+data.getString("key"));
            entities.add(new Sign(data.getFloat("x"), data.getFloat("y"), msgs, data.getFloat("img_scale"), data.getFloat("text_size")));
            spawned = true;
          } else if (data.getString("sign_type").equals("image")) {
            entities.add(new Sign(data.getFloat("x"), data.getFloat("y"), data.getString("img_name"), data.getFloat("display_img_scale"), data.getFloat("img_scale")));
            spawned = true;
          }
        } else if (type.equals("rain_cloud")) {
          entities.add(new RainCloud(data.getFloat("x"), data.getFloat("y"), data.getFloat("xv"), data.getFloat("yv"), data.getFloat("img_scale"), data.getFloat("interval"), data.getBoolean("can_follow")));
          spawned = true;
        } else if (type.equals("running_checkpoint")) {
          entities.add(new RunningCheckPoint(data.getFloat("x"), data.getFloat("y"), data.getFloat("dist")));
          spawned = true;
        } else if (type.equals("portal")) {
          entities.add(new Portal(data.getFloat("x1"), data.getFloat("y1"), data.getFloat("x2"), data.getFloat("y2")));
          spawned = true;
        }
        
        if (!data.isNull("id") && spawned) {
          entities.get(entities.size()-1).id = data.getFloat("id");
        }
      }
    }
    
    if (keep_player_location) {
      game.player = new Player(game.player.checkpoint.x, game.player.checkpoint.y);
    } else {
      game.player = new Player(0, 0);
      //game.player = new Player(8580, 30);
      //game.player = new Player(630, 600);
      //game.player = new Player(330, 990);
      //game.player = new Player(4920, 2445);
      //game.player = new Player(5700, 2955);
      //game.player = new Player(7530, 4245);
      //game.player = new Player(7590, 4620);
    }
    return new Level(blocks, bullets, enemies, entities);
  }
  
  Block get_block(String t, float x, float y) {
    /*
    @ - Normal block
    # - Disappearing block
    $ - Appearing block
    i - invisible block
    wasd - Spike block
    rty - Spring block
    WASD - Shooter block
    qQ - Dodging block (always proximity, varied recurrent) q - false, Q - true
    c - crash block
    m - MSOD
    v - poly-collision block
    pP - piston block (P inverted)
    b - barrel block
    */
    
    if (t.equals("")) {
      return null;
    } else if (t.equals("@")) {
      return new NormalBlock(x, y);
    } else if (t.equals("#")) {
      return new DisappearingBlock(x, y);
    } else if (t.equals("$")) {
      return new AppearingBlock(x, y);
    } else if (t.equals("i")) {
      Block block = new NormalBlock(x, y);
      block.visible = false;
      return block;
    } else if ("wasd".indexOf(t) != -1) {
      return new SpikeBlock(x, y, t);
    } else if ("rty".indexOf(t) != -1) {
      return new SpringBlock(x, y, t, 25);
    } else if ("WASD".indexOf(t) != -1) {
      return new ShooterBlock(x, y, t);
    } else if ("qQ".indexOf(t) != -1) {
      return new DodgingBlock(x, y, t.equals("Q"), true);
    } else if (t.equals("c")) {
      return new GameCrashBlock(x, y);
    } else if (t.equals("m")) {
      return new MSOD(x, y);
    } else if (t.equals("v")) {
      return new PolyCollisionBlock(x, y);
    } else if ("pP".indexOf(t) != -1) {
      return new PistonBlock(x, y, 60, t.equals("P"));
    } else if (t.equals("b")) {
      return new BarrelBlock(x, y);
    }
    return null;
  }
  
  Enemy get_enemy(String type, float x, float y) {
    /*
    1 - Happy face
    2 - Troll farmer
    3 - Flying cat
    4 - Frog
    5 - Fire breathing frog
    */
    int t = int(type);
    
    if (t == 1) {
      return new HappyFace(x, y);
    } else if (t == 2) {
      return new TrollFarmer(x, y);
    } else if (t == 3) {
      return new FlyingCat(x, y);
    } else if (t == 4) {
      return new Frog(x, y);
    } else if (t == 5) {
      return new FireBreathingFrog(x, y);
    }
    
    return null;
  }
}