class Button {
  PVector location, rectSize;
  String img_name, hovered_img_name, text;
  float text_size, img_scale;
  boolean clicked, text_mode;
  color clr1, clr2;

  Button(float x, float y, String _img_name, float _img_scale) {
    location = new PVector(x, y);
    img_name = _img_name;
    clicked = false;
    rectSize = new PVector(gt(img_name).width, gt(img_name).height);
    text_mode = false;
    img_scale = _img_scale;
  }
  
  Button(float x, float y, String _img_name, String _hovered_img_name, float _img_scale) {
   location = new PVector(x, y);
    img_name = _img_name;
    hovered_img_name = _hovered_img_name;
    clicked = false;
    rectSize = new PVector(gt(img_name).width, gt(img_name).height);
    text_mode = false;
    img_scale = _img_scale;
  }
  
  Button(float x, float y, String _text, float _text_size, color _clr1) {
    location = new PVector(x, y);
    text = _text;
    text_size = _text_size;
    clicked = false;
    textSize(text_size);
    rectSize = new PVector(textWidth(_text), text_size);
    clr1 = _clr1;
    clr2 = _clr1;
    text_mode = true;
  }
  
  Button(float x, float y, String _text, float _text_size, color _clr1, color _clr2) {
   location = new PVector(x, y);
    text = _text;
    text_size = _text_size;
    clicked = false;
    textSize(text_size);
    rectSize = new PVector(textWidth(_text), text_size);
    clr1 = _clr1;
    clr2 = _clr2;
    text_mode = true;
  }
  
  void update() {
    //if (!clicked) {
      display();
    //}
  }

  void display() {
    if (text_mode) {
      textAlign(CENTER, CENTER);
      textSize(text_size);
      fill(clr1);
      if (mouseOn()) {
        fill(clr2);
      }
      text(text, location.x, location.y);
    } else {
      pushMatrix();
      translate(location.x, location.y);
      scale(img_scale);
      if ((hovered_img_name != null) && mouseOn()) {
        image(texture_manager.get_texture(hovered_img_name), 0, 0);
      } else {
        image(texture_manager.get_texture(img_name), 0, 0);
      }
      popMatrix();
    }
  }

  boolean mouseOn () {
    if ((mouseX > location.x - rectSize.x/2.5) &&
        (mouseX < location.x + rectSize.x/2.5) &&
        (mouseY > location.y - rectSize.y/2.5) &&
        (mouseY < location.y + rectSize.y/2.5)) {
      return true;
    } else {
      return false;
    }
  }
  
  void mouseClicked() {
    if (mouseOn()) {
      clicked = true;
    }
  }
}

class RunningButton extends Button {
  PVector destination, lerp_factor, original_location;
  boolean triggered, moving;
  int runs, max_runs;
  RunningButton(float x, float y, String _text, float _text_size, color _clr1) {
    super(x, y, _text, _text_size, _clr1);
    destination = new PVector(x, y);
    original_location = new PVector(x, y);
    moving = false;
    runs = 0;
    max_runs = 5;
    lerp_factor = new PVector(0, 0);
    triggered = false;
  }
  
  RunningButton(float x, float y, String img_name, float img_scale) {
    super(x, y, img_name, img_scale);
    destination = new PVector(x, y);
    original_location = new PVector(x, y);
    moving = false;
    runs = 0;
    max_runs = 5;
    lerp_factor = new PVector(0, 0);
    triggered = false;
  }
  
  void update() {
    move();
    super.update();
  }
  
  void move() {
    if (!triggered) {
      return;
    }
    if (moving) {
      //location.lerp(destination, 0.4);
      location.x = lerp(location.x, destination.x, lerp_factor.x);
      location.y = lerp(location.y, destination.y, lerp_factor.y);
      if (abs(location.x-destination.x) <= 25 && abs(location.y-destination.y) <= 25) {
        runs++;
        moving = false;
      }
    }
    
    if (!moving && mouseOn()) {
      moving = true;
      if (runs < max_runs-1) {
        destination = new PVector(random(rectSize.x/2, width-rectSize.x/2), random(rectSize.y/2, height-rectSize.y/2));
        game.sound_engine.gs("nope").play();
      } else if (runs < max_runs) {
        destination = original_location.copy();
        game.sound_engine.gs("nope").play();
      }
      float lerp_1 = ((int)random(1)*4+3)*0.1;
      lerp_factor = new PVector(lerp_1, 1-lerp_1);
    }
  }
  
  void mouseClicked() {
    if (!triggered) {
      triggered = true;
    } else if (mouseOn()) {
      clicked = true;
    }
  }
}

class Text {
  float setup_counter, typing_counter;
  boolean is_dead;
  int text_index, max_typing_counter;
  String[] text, names, face_img_names;
  float [] face_img_scales;
  int[] order;
  
  Text (String[] _text, String[] _names, String[] _face_img_names, float[] _face_img_scales, int[] _order) {
    text = _text;
    names = _names;
    face_img_names = _face_img_names;
    face_img_scales = _face_img_scales;
    order = _order;
    
    is_dead = false;
    text_index = 0;
    setup_counter = 0;
    typing_counter = 0;
    max_typing_counter = text[0].length();
  }
  
  void update() {
    display();
  }
  
  void display() {
    if (is_dead) {
      return;
    }
    if (text_index >= text.length) {
      if (setup_counter > 0) {
        setup_counter = constrain(setup_counter, 0, width);
        
        fill(0);
        rectMode(CORNER);
        rect(width-setup_counter, height*0.7, width, height*0.3);
        rectMode(CENTER);
        
        fill(255);
        //image(texture_manager.get_texture(face_img_name), height*0.15-width/3+setup_counter/3, height*0.85);
        display_texture(height*0.15-width/3+setup_counter/3, height*0.85, face_img_names[order[text_index-1]], face_img_scales[order[text_index-1]]);
        
        setup_counter -= game.dt*30;
      } else {
        text_index = 0;
        is_dead = true;
      }
      return;
    }
    
    if (setup_counter < width) {
      setup_counter = constrain(setup_counter, 0, width);
      
      fill(0);
      rectMode(CORNER);
      rect(width-setup_counter, height*0.7, width, height*0.3);
      rectMode(CENTER);
      
      fill(255);
      //image(texture_manager.get_texture(face_img_name), height*0.15-width/3+setup_counter/3, height*0.85);
      display_texture(height*0.15-width/3+setup_counter/3, height*0.85, face_img_names[order[text_index]], face_img_scales[order[text_index]]);
      
      setup_counter += game.dt*30;
      return;
    }
    
    fill(0);
    rectMode(CORNER);
    rect(0, height*0.7, width, height*0.3);
    rectMode(CENTER);
    
    fill(255);
    //image(texture_manager.get_texture(face_img_name), height*0.15, height*0.85);
    display_texture(height*0.15, height*0.85, face_img_names[order[text_index]], face_img_scales[order[text_index]]);
    
    textSize(25);
    textAlign(LEFT, TOP);
    fill(255);
    textFont(game.sign_font);
    //text(name + ":\n" + text.get(text_index), height*0.40, height*0.75);
    if (names[order[text_index]].equals("PlayerName")) {
      text(game.player_name + ":", height*0.40, height*0.75);
    } else {
      text(names[order[text_index]] + ":", height*0.40, height*0.75);
    }
    text(text[text_index].substring(0, round(typing_counter)), height*0.40, height*0.80);
    
    typing_counter = constrain(typing_counter+game.dt/3, 0, max_typing_counter);
  }
  
  void display_texture(float x, float y, String img_name, float img_scale) {
    pushMatrix();
    translate(x, y);
    scale(img_scale);
    image(gt(img_name), 0, 0);
    popMatrix();
  }
  
  void mouseClicked() {
    if (text_index >= text.length) {
      return;
    }
    if (typing_counter < max_typing_counter) {
      typing_counter = max_typing_counter;
    } else {
      text_index++;
      if (text_index > text.length) {
        text_index = text.length - 1;
        return;
      }
      //typing_counter = 0;
      //if (text_index < text.length) {
      //  max_typing_counter = text[text_index].length();
      //}
      
      if (text_index < text.length && text[text_index].length() >= 7 && text[text_index].substring(text[text_index].length()-7, text[text_index].length()).equals("_cactus")) {
        text[text_index] = text[text_index].substring(0, text[text_index].length()-7);
        
        for (HashMap.Entry<PVector,Block> entry : game.blocks.entrySet()) {
          Block block = entry.getValue();
          if (block.getClass().getSimpleName().equals("CactusBlock")) {
            entry.getValue().properties.setFloat("cactus_transition", 1); 
          }
        }
        game.sound_engine.gs("cactus").play();
      }
      
      if (text_index < text.length && text[text_index].length() >= 9 && text[text_index].substring(text[text_index].length()-9, text[text_index].length()).equals("_teleport")) {
        text[text_index] = text[text_index].substring(0, text[text_index].length()-9);
        
        for (Entity entity : game.entities) {
          if (entity.getClass().getSimpleName().equals("BoatMan")) {
            game.player.location = new PVector(entity.location.x+200, entity.location.y-32);
          }
        }
      }
      
      typing_counter = 0;
      if (text_index < text.length) {
        max_typing_counter = text[text_index].length();
      }
    }
  }
}

class LolText {
  PVector location;
  color theColor;
  LolText(float x, float y) {
    location = new PVector(x, y);
    theColor = color(random(255), random(255), random(255));
  }
  
  void display() {
    textAlign(CENTER, CENTER);
    textSize(30);
    fill(theColor);
    text(game.game_over_text, location.x, location.y);
  }
}

class SoundEngine {
  //loads a bunch of sound files that can be referenced and played later.
  Minim minim;
  //AudioPlayer crickets, screech, theme, duck;
  HashMap<String,AudioPlayer> sounds;
  ArrayList<AudioPlayer> temp_sounds;

  SoundEngine(PApplet p) {
    minim = new Minim(p);
    sounds = new HashMap<String,AudioPlayer>();
    temp_sounds = new ArrayList<AudioPlayer>();
    loadSounds();
    
    //crickets.setGain(-30);
    sounds.get("crickets").setGain(-30);
    
    //mute code
    //for (Map.Entry<String, AudioPlayer> sound : sounds.entrySet()) {
    //  sound.getValue().setGain(-80);
    //}
  }
  
  void update() {
    //for (AudioPlayer player : sounds.values()) {
    //  if (player.position() == player.length()) {
    //    player.rewind();
    //  }
    //}
    for (Map.Entry<String, AudioPlayer> sound : sounds.entrySet()) {
      if (!sound.getKey().equals("theme") && sound.getValue().position() == sound.getValue().length()) {
        sound.getValue().rewind();
      }
    }
  }
  
  void loadSounds() {
    String path = sketchPath();
    ArrayList<File> allFiles = listFilesRecursive(path);
    String[] extentions = {".wav", ".mp3", ".ogg"};
    for (File f : allFiles) {
      if (Arrays.asList(extentions).contains(f.getAbsolutePath().substring(f.getAbsolutePath().length()-4))) {
        String name = f.getAbsolutePath();
        sounds.put(splitString(name).split("\\.")[0], minim.loadFile(name));
      }
    }
  }
  
  String splitString(String name) {
    String f_name = "error";
    for (int i = name.length()-1; i >= 0; i--) {
      if (name.charAt(i) == '/' || name.charAt(i) == '\\') {
        f_name = name.substring(i+1);
        break;
      }
    }
    return f_name;
  }
  
  AudioPlayer gs(String sound_name) {
    return sounds.get(sound_name.split("\\.")[0]);
  }
  
  ArrayList<File> listFilesRecursive(String dir) {
    ArrayList<File> fileList = new ArrayList<File>(); 
    recurseDir(fileList, dir);
    return fileList;
  }
  
  void recurseDir(ArrayList<File> a, String dir) {
    File file = new File(dir);
    if (file.isDirectory()) {
      a.add(file);  
      File[] subfiles = file.listFiles();
      for (int i = 0; i < subfiles.length; i++) {
        recurseDir(a, subfiles[i].getAbsolutePath());
      }
    } else {
      a.add(file);
    }
  }
}