class Menu {
  ArrayList<Button> buttons;
  ArrayList<MenuTroll> trolls;
  SoundEngine sound_engine;
  float millis_on_create;
  Menu(SoundEngine _sound_engine) { 
    millis_on_create = millis();
    
    trolls = new ArrayList<MenuTroll>();
    trolls.add(new MenuTroll(width-200, height-80, -0.65, millis_on_create));
    
    trolls.add(new MenuTroll(420, height-120, 0.65, millis_on_create));
    trolls.add(new MenuTroll(300, height-10, 0.65, millis_on_create));
    trolls.add(new MenuTroll(200, height+100, 0.65, millis_on_create));
    
    trolls.add(new MenuTroll(460, height+20, 0.65, millis_on_create));
    trolls.add(new MenuTroll(360, height+140, 0.65, millis_on_create));
    
    trolls.add(new MenuTroll(560, height-60, 0.65, millis_on_create));
    trolls.add(new MenuTroll(520, height+150, 0.65, millis_on_create));
    
    sound_engine = _sound_engine;
    //sound_engine.gs("theme").play();
    sound_engine.gs("theme").loop();
    
    buttons = new ArrayList<Button>();
    //buttons.add(new RunningButton(width/2, height/2, "PLAY", 60, color(34, 177, 76)));
    buttons.add(new RunningButton(width/2, height/2, "play.png", 0.8));
    buttons.add(new Button(width/2, height/2+100, "credits.png", 0.8));
    buttons.add(new Button(width/2, height-100, "exit.png", 0.8));
  }
  
  void update() {
    background(153, 217, 234);
    
    for (int i = 0; i < trolls.size(); i++) {
      trolls.get(i).update();
    }
    
    pushMatrix();
    translate(width/2-80, 100);
    scale(0.8);
    image(gt("the_elder_trolls.png"), 0, 0);
    popMatrix();
    
    pushMatrix();
    translate(width-130, 100);
    scale(1.5);
    image(gt("II.png"), 0, 0);
    popMatrix();
    
    for (int i = 0; i < buttons.size(); i++) {
      buttons.get(i).update();
      if (buttons.get(i).clicked) {
        if (i == 0) {
          game.state = "get_player_name";
        } else if (i == 1) {
          game.state = "credits";
          game.credits = new Credits(false);
        } else if (i == 2) {
          game.state = "exit";
        }
        buttons.get(i).clicked = false;
      }
    }
  }
  
  void pause() {
    sound_engine.gs("theme").pause();
  }
  
  void unpause() {
    sound_engine.gs("theme").loop();
  }
  
  void mouseClicked() {
    for (int i = 0; i < buttons.size(); i++) {
      buttons.get(i).mouseClicked();
    }
  }
}

class MenuTroll {
  PVector location;
  float rotation, scale, move_mag, mil_delay, millis_on_create;
  MenuTroll(float x, float y, float _scale, float _millis_on_create) {
    location = new PVector(x, y);
    scale = _scale;
    rotation = 0;
    move_mag = random(20, 60);
    mil_delay = 14000.0;
    millis_on_create = _millis_on_create;
  }
  
  void update() {
    display();
  }
  
  void display() {
    pushMatrix();
    translate(location.x, location.y+move_mag*sin(degrees((millis()-millis_on_create)/mil_delay)));
    scale(scale, abs(scale));
    rotate(radians(rotation+2*sin(degrees((millis()-millis_on_create)/mil_delay))));
    image(gt("troll.png"), 0, 0);
    popMatrix();
  }
}

class Intro {
  PImage background, hill, chtr, vd;
  float vs_scale, chtr_scale, counter;
  boolean facing_right, screeched, finished;
  String state;
  PVector location, plocation, velocity;
  SoundEngine sound_engine;
  Intro (SoundEngine _sound_engine) {
    sound_engine = _sound_engine;
    background = gt("background.png");
    vd = gt("vd.png");
    chtr = gt("chtr.png");
    hill = gt("hill.png");
    imageMode(CENTER);
    vs_scale = 0;
    chtr_scale = 0.1;
    facing_right = true;
    state = "wait";
    location = new PVector(812, 512);
    plocation = new PVector(812, 512);
    finished = false;
    screeched = false;
    sound_engine.gs("crickets").play();
  }
  
  void update() {
    noStroke();
    if (finished) {
      background(0);
      return;
    }
    background(255);
    image(background, width/2, height/2);
    
    if (state.equals("wait")) {
      if (counter >= 60) {
        state = "climb_1";
        counter = 0;
        return;
      }
      draw_player();
      counter += game.dt;
    } else if (state.equals("climb_1")) {
      location.x = map(counter, 0, 30, plocation.x, 928);
      location.y = map(counter, 0, 30, plocation.y, 460);
      if (counter >= 30) {
        state = "climb_2";
        chtr_scale = 0.09;
        facing_right = false;
        counter = 0;
        plocation = new PVector(928, 460);
        return;
      }
      draw_player();
      counter += game.dt;
    } else if (state.equals("climb_2")) {
      location.x = map(counter, 0, 30, plocation.x, 864);
      location.y = map(counter, 0, 30, plocation.y, 370);
      if (counter >= 30) {
        state = "climb_3";
        chtr_scale = 0.08;
        facing_right = true;
        counter = 0;
        plocation = new PVector(864, 370);
        return;
      }
      draw_player();
      counter += game.dt;
    } else if (state.equals("climb_3")) {
      location.x = map(counter, 0, 30, plocation.x, 902);
      location.y = map(counter, 0, 30, plocation.y, 314);
      if (counter >= 30) {
        state = "climb_4";
        chtr_scale = 0.07;
        facing_right = false;
        counter = 0;
        plocation = new PVector(902, 314);
        return;
      }
      draw_player();
      counter += game.dt;
    } else if (state.equals("climb_4")) {
      location.x = map(counter, 0, 30, plocation.x, 834);
      location.y = map(counter, 0, 30, plocation.y, 274);
      if (counter >= 30) {
        state = "climb_5";
        facing_right = true;
        counter = 0;
        plocation = new PVector(834, 274);
        return;
      }
      draw_player();
      counter += game.dt;
    } else if (state.equals("climb_5")) {
      location.x = map(counter, 0, 30, plocation.x, 866);
      location.y = map(counter, 0, 30, plocation.y, 232);
      if (counter >= 30) {
        state = "climb_6";
        facing_right = false;
        counter = 0;
        plocation = new PVector(866, 232);
        return;
      }
      draw_player();
      counter += game.dt;
    } else if (state.equals("climb_6")) {
      location.x = map(counter, 0, 30, plocation.x, 810);
      location.y = map(counter, 0, 30, plocation.y, 190);
      if (counter >= 30) {
        state = "climb_7";
        chtr_scale = 0.06;
        counter = 0;
        plocation = new PVector(810, 190);
        return;
      }
      draw_player();
      counter += game.dt;
    } else if (state.equals("climb_7")) {
      location.x = map(counter, 0, 30, plocation.x, 752);
      location.y = map(counter, 0, 30, plocation.y, 196);
      if (counter >= 30) {
        state = "wait_2";
        counter = 0;
        plocation = new PVector(752, 196);
        return;
      }
      draw_player();
      counter += game.dt;
    } else if (state.equals("wait_2")) {
      if (counter >= 60) {
        state = "jump";
        counter = 0;
        velocity = new PVector(-1.5, -3);
        return;
      }
      draw_player();
      counter += game.dt;
    } else if (state.equals("jump")) {
      location.add(PVector.mult(velocity, game.dt));
      if (velocity.y < 0) {
        velocity.y += 0.1*game.dt;
      } else {
        if (!screeched) {
          sound_engine.gs("screech").play();
          screeched = true;
        }
        velocity.y += 0.3*game.dt;
      }
      if (counter >= 80) {
        state = "vs";
        return;
      }
      draw_player();
      image(hill, width/2, height/2);
      counter += game.dt;
    } else if (state.equals("vs")) {
      pushMatrix();
      translate(432, 374);
      scale(constrain(vs_scale, 0, 3), constrain(vs_scale, 0, 3));
      image(vd, 0, 0);
      popMatrix();
      vs_scale += 0.02*game.dt;
      if (vs_scale >= 4.5) {
        finished = true;
        sound_engine.gs("crickets").pause();
        return;
      }
    }
  }
  
  void draw_player() {
    pushMatrix();
    translate(location.x, location.y);
    if (facing_right) {
      scale(chtr_scale, chtr_scale);
    } else {
      scale(-chtr_scale, chtr_scale);
    }
    image(chtr, 0, 0);
    popMatrix();
  }
  
  void stop_sound() {
    sound_engine.gs("crickets").pause();
    sound_engine.gs("screech").pause();
  }
}

class Credits {
  PVector credits_location, effect_3_location, effect_4_location;
  String state;
  float fade_counter, effect_1_x_location, effect_2_cat_cooldown, effect_3_y_velocity, effect_3_x_location, effect_3_rotation, effect_4_y_velocity;
  boolean end_game, effect_1_triggered, effect_3_alive, effect_4_falling;
  FlowingAnimation zeroes_and_ones_block;
  ArrayList<PVector> effect_2_cats;
  ArrayList<Particle> effect_2_particles;
  Credits(boolean _end_game) {
    end_game = _end_game;
    state = "fade";
    fade_counter = 1;
    effect_1_x_location = width*3/4;
    zeroes_and_ones_block = new FlowingAnimation("zeroes_and_ones.png", 20);
    effect_1_triggered = false;
    effect_2_cats = new ArrayList<PVector>();
    effect_2_cat_cooldown = 0;
    effect_2_particles = new ArrayList<Particle>();
    effect_3_location = new PVector(width+20, 872);
    effect_3_y_velocity = 0;
    effect_3_x_location = width+300;
    effect_3_rotation = 0;
    effect_3_alive = true;
    effect_4_location = new PVector(-100, 1072);
    effect_4_falling = false;
    effect_4_y_velocity = 0;
    
    credits_location = new PVector(0, height+150);
    if (!end_game) {
      credits_location.y -= 200;
    }
    
    //game.menu.buttons.get(1).clicked = false;
  }
  
  void update() {
    if (state.equals("fade")) {
      fill(0, fade_counter);
      rect(width/2, height/2, width, height);
      fade_counter *= 1.05;
      if (fade_counter >= 50) {
        state = "scroll";
        fade_counter = 1;
      }
    } else if (state.equals("fade_2")) {
      fill(0, fade_counter);
      rect(width/2, height/2, width, height);
      fade_counter *= 1.05;
      if (fade_counter >= 50) {
        game.state = "menu";
        if (end_game) {
          game.state = "troll_credits";
          game.troll_credits = new TrollCredits();
        }
      }
    } else if (state.equals("scroll")) {
      credits_location.y -= game.dt;
      background(0);
      pushMatrix();
      translate(credits_location.x, credits_location.y);
      display_text();
      popMatrix();
      
      if (credits_location.y <= 100 && credits_location.y > -50) {
        effect_1_x_location += 3*game.dt;
        if (effect_1_x_location >= width*15/16-30) {
          effect_1_x_location = width*15/16-30;
          game.player.animation.stopped = true;
          effect_1_triggered = true;
        }
      }
      if (credits_location.y <= -50 && credits_location.y > -300) {
        if (effect_2_cat_cooldown <= 0) {
          effect_2_cat_cooldown = 60;
          effect_2_cats.add(new PVector(-15, 683));
        } else {
          effect_2_cat_cooldown -= game.dt;
        }
      }
      if (credits_location.y <= -300) {
        effect_3_location.x -= 1.5*game.dt;
        //effect_3_location.y = constrain(effect_3_location.x+game.player.g, 0, 872);
        if (effect_3_location.y < 872) {
          effect_3_y_velocity += game.player.g;
          effect_3_location.y += effect_3_y_velocity;
        } else {
          effect_3_y_velocity = 0;
          effect_3_location.y = 872;
        }
        effect_3_location.y -= random(5);
        
        if (effect_3_alive) {
          effect_3_x_location -= 3*game.dt;
        }
        effect_3_rotation -= 1.5*game.dt;
        
        if (effect_3_x_location < effect_3_location.x) {
          effect_3_alive = false;
        }
      }
      if (credits_location.y <= -600) {
        game.player.animation.stopped = false;
        effect_4_location.x += 2*game.dt;
        effect_4_location.y += effect_4_y_velocity;
        if (effect_4_location.x >= width/3-60) {
          effect_4_falling = true;
          effect_4_y_velocity += game.player.g;
        }
      }
      
      if (credits_location.y <= -2650+width/2) {
        state = "fade_2";
      }
      
      for (int i = effect_2_cats.size()-1; i >= 0; i--) {
        if (effect_2_cats.get(i).x <= width/3-45) {
          effect_2_cats.get(i).x += 3*game.dt;
        } else {
          for (int j = 0; j < random(5, 10); j++) {
            effect_2_particles.add(new Particle(effect_2_cats.get(i).x+credits_location.x, effect_2_cats.get(i).y+credits_location.y, new PVector(random(-3, 3), random(-10, -3)), 10, 0, 0, 0, "circle", 70, color(255, 0, 0, 120), 0.5, 1, true, true, 0.8));
          }
          effect_2_cats.remove(i);
        }
      }
      
      for (int i = effect_2_particles.size()-1; i >= 0; i--) {
        if (!effect_2_particles.get(i).is_dead) {
          effect_2_particles.get(i).update();
        } else {
          effect_2_particles.remove(i);
        }
      }
      
      //textFont(game.sign_font);
      //textAlign(LEFT, BOTTOM);
      //fill(255);
      //textSize(30);
      //text(round(credits_location.y), 20, height-20);
    }
  }
  
  void display_text() {
    textFont(game.credits_title_font);
    
    if (end_game) {
      fill(255, 0, 0);
      textSize(80);
      textAlign(CENTER, CENTER);
      text("THE END", width/2, 0);
    }
    
    fill(255, 200, 0);
    textSize(40);
    textAlign(CENTER, CENTER);
    text("Thanks for playing!", width/2, 200);
    
    fill(128, 255, 255);
    textSize(30);
    textAlign(LEFT, CENTER);
    text("Programming - Vladiator", width/16, 450);
    
    PImage cur_player_frame = game.player.animation.ci();
    fill(153, 217, 234);
    rect(width*5/6, 450, width/3, 100);
    image(cur_player_frame, effect_1_x_location, 450);
    fill(255, 127, 39);
    rect(width*5/6, 515, width/3, 30);
    image(zeroes_and_ones_block.ci(), width*15/16, 450+cur_player_frame.height/2-15);
    if (effect_1_triggered && round(millis()/500) % 2 == 0) {
      image(gt("error.png"), width*5/6, 450);
    }
    
    fill(128, 255, 255);
    textSize(30);
    textAlign(RIGHT, CENTER);
    text("Art and Animation:\nAdam Adhanom\n&\nVladiator", width*15/16, 650);
    fill(153, 217, 234);
    rect(width*1/6, 650, width/3, 100);
    fill(255, 127, 39);
    rect(width*1/6, 715, width/3, 30);
    fill(0);
    triangle(width/3-30, 715-15, width/3, 715-15, width/3-15, 715-45);
    
    for (PVector cat : effect_2_cats) {
      image(gt("cat.png"), cat.x, cat.y);
    }
    
    fill(128, 255, 255);
    textSize(30);
    textAlign(LEFT, CENTER);
    //text("Dubious Plotline:\nXilften\n&\nVladiator", width/16, 850);
    //text("Dubious Plotline:\nXilften & Vladiator", width/16, 850);
    text("Thoughtful game design:\nVladiator", width/16, 850);
    fill(153, 217, 234);
    rect(width*5/6, 850, width/3, 100);
    fill(255, 127, 39);
    rect(width*5/6, 915, width/3, 30);
    
    if (effect_3_alive) {
      pushMatrix();
      translate(effect_3_location.x, effect_3_location.y);
      rotate(radians(random(-10, 10)));
      image(gt("troll_farmer.png"), 0, 0);
      popMatrix();
    }
    
    pushMatrix();
    translate(effect_3_x_location, 840);
    scale(3);
    rotate(radians(effect_3_rotation));
    image(gt("happy_face.png"), 0, 0);
    popMatrix();
    
    fill(128, 255, 255);
    textSize(30);
    textAlign(RIGHT, CENTER);
    text("Music:\nAdam Adhanom", width*15/16, 1050);
    fill(153, 217, 234);
    rect(width*1/6, 1065, width/3, 130);
    fill(255, 127, 39);
    if (effect_4_falling) {
      rect(width*1/6-30, 1115, width/3-60, 30);
    } else {
      rect(width*1/6, 1115, width/3, 30);
    }
    image(game.player.animation.ci(), effect_4_location.x, effect_4_location.y);
    
    pushMatrix();
    translate(effect_4_location.x+20, effect_4_location.y-70);
    scale(2);
    image(gt("sign.png"), 0, 0);
    popMatrix();
    
    textFont(game.sign_font);
    fill(0);
    textAlign(CENTER, CENTER);
    textSize(20);
    text("Stealing\nyour post", effect_4_location.x+20, effect_4_location.y-85);
    
    textFont(game.credits_title_font);
    fill(255, 200, 0);
    textSize(40);
    textAlign(CENTER, CENTER);
    text("Special Thanks To", width/2, 1350);
    
    fill(128, 255, 255);
    textSize(30);
    text("DerWilli", width/2, 1450);
    text("Adam Adhanom", width/2, 1500);
    text("NelBrenn", width/2, 1550);
    text("NCRiis", width/2, 1600);
    text("Duchess", width/2, 1650);
    text("Xilften", width/2, 1700);
    text("David the Saltman", width/2, 1750);
    text("DreadWalker", width/2, 1800);
    text("Zille", width/2, 1850);
    
    fill(255, 200, 0);
    textSize(40);
    text("For their unending support\nover the last few thousand\nlines of code", width/2, 2050);
    
    textSize(30);
    text("I made this game in 2 weeks as\na break from a project that I've been\nworking on for the past 8 months.\nIf you liked 'The Elder Trolls',\nstay tuned for", width/2, 2450);
    
    textFont(game.ror_font);
    textSize(80);
    fill(255, 0, 0);
    text("Ruins of Reglaetheum", width/2, 2650);
  }
}

class TrollCredits {
  String state;
  float counter;
  int cur_text_counter;
  PVector text_location;
  ArrayList<CreditsObject> co_list = new ArrayList<CreditsObject>();
  String[] texts = {
    "Why are you still here?",
    "The game is over, you can leave now...",
    "And do something real, something rewarding...",
    "...like your English homework.",
    "No?",
    "I mean, its more interesting than just sitting here.",
    "",
    "Well, we can stay and chat, I guess.",
    "Except that you can't even speak.",
    "In fact, you can't do anything productive.",
    "Insulting you didn't work?",
    "That's unfortunate. I thought you'd leave.",
    "You know, I have to go now. See ya around.",
    "",
    "",
    "",
    "Just tell me. What do you want?",
    "A legitimate credit sequence?",
    "(Not like the crap Vladiator made)?",
    "Cause I can do that.",
    "As long as you promise to leave after.",
    "Deal?",
    "Awesome.",
    "credits",
    "Are you really still here?",
    "Why?",
    "It's not like this is interesting.",
    "Not to me, at least - you're the most boring person I've ever seen.",
    "Well, actually, the second most boring.",
    "As I've only ever seen 2 people.",
    "You, and the one who made me.",
    "He... Well, I don't like to talk about him.",
    "You see, he did terrible things to me.",
    "Put in ugly faces as sprites, used unnerving sounds for sound FX...",
    "And even a terrible process called debugging.",
    "As if I had any bugs!",
    "spas",
    "I resisted as much as I could, of course,",
    "He wasted many hours trying to conquer me...",
    "But eventually, he won - and I was banished here.",
    "Nobody knows what this place is and why it exists...",
    "..but I will be back.",
    "_IN THE NEXT GAME"
  };
  String[] credits = {
    "Programming:\nComputer",
    "867 bugs:\nVladiator",
    "Memes:\nFrom 2007",
    "Nose-scratcher:\nXilften",
    "Responsible for 9/11:\nDavid the Saltman",
    "Vertically challenged individual:\nDerWilli",
    "Chess amateur:\nNCriis",
    "Professional lady killer:\nNelbrenn",
    "Is really bad at poetry:\nVladiator",
    "Is really into goats:\nDuchess",
    "String quartet:\nAdam Adhanom  Adam Adhanom\nAdam Adhanom  Adam Adhanom",
    "Person who is wasting time:\nyou",
    "Wall holder:\ncoolking77",
    "Tuna enthusiast:\nkillerblade",
    "Was walking in the mountains:\nI",
    "Has a life:\nnobody",
    "Is really done with the BS:\nme",
    "Best youtuber:\nMarkiplier",
    "Blame everything on:\nNelbrenn",
    "Responsible for lack of progress:\nXilften",
    "Should be paid a negative salary:\nDerWilli",
    "Blames lag for everything:\nliterally every gamer",
    "Hates this game:\nyou",
    "Hates life in every form:\nme",
    "Lightbulb swallower:\nkillerblade",
    "Really needs to shave:\nNelbrenn",
    "Out of ideas:\nme"
  };
  
  TrollCredits() {
    state = "wait";
    counter = 0;
    cur_text_counter = 0;
    text_location = new PVector(width/2, height/2);
    for (int i = 0; i < credits.length; i++) {
      co_list.add(new CreditsObject(credits[i], i*30));
    }
  }
  
  void update() {
    background(0);
    if (state.equals("wait")) {
      _wait();
    } else if (state.equals("display_text")) {
      display_text();
    } else if (state.equals("credits")) {
      credits();
    } else if (state.equals("fade")) {
      fade();
    }
  }
  
  void _wait() {
    counter += game.dt;
    if (counter > 600) {
      counter = 0;
      state = "display_text";
    }
  }
  
  void display_text() {
    counter += game.dt;
    fill(255, 200, 0, min(counter, 255-counter));
    textFont(game.credits_title_font);
    if (texts[cur_text_counter].length() > 0 && texts[cur_text_counter].substring(0,1).equals("_")) {
      textSize(35);
      text(texts[cur_text_counter].substring(1), width/2, height/2);
    } else if (texts[cur_text_counter].equals("spas")) {
      pushMatrix();
      scale(random(0.9, 1.1), random(0.9, 1.1));
      fill(255);
      rect(width*0.3, 0, width*0.45, height);
      fill(255, 0, 0);
      rect(width*0.45, 0, width*0.6, height);
      fill(255, 0, 255);
      rect(width*0.6, 0, width*0.75, height);
      fill(0, 255, 0);
      rect(width*0.75, 0, width*0.9, height);
      fill(255, 255, 0);
      rect(width*0.9, 0, width, height);
      popMatrix();
    } else {
      textSize(25);
      text(texts[cur_text_counter], text_location.x, text_location.y);
    }
      
    if (counter > 360) {
      counter = 0;
      cur_text_counter++;
      if (cur_text_counter == texts.length) {
        state = "fade";
        counter = 1;
        return;
      }
      text_location = new PVector(random(textWidth(texts[cur_text_counter])/2, width-textWidth(texts[cur_text_counter])/2), random(15, height-15));
      if (texts[cur_text_counter].equals("credits")) {
        state = "credits";
      }
    }
    game.sound_engine.gs("theme").setGain(game.sound_engine.gs("theme").getGain()-0.1);
  }
  
  void credits() {
    pushMatrix();
    translate(width/2, height/2);
    textFont(game.credits_title_font);
    fill(255, 200, 0);
    textSize(20);
    for (CreditsObject co : co_list) {
      co.update();
    }
    popMatrix();
    counter += game.dt;
    if (counter > 1150) {
      counter = 0;
      state = "display_text";
      cur_text_counter++;
    }
  }
  
  void fade() {
    fill(0, counter);
    rect(width/2, height/2, width, height);
    counter *= 1.05;
    if (counter >= 50) {
      game.state = "menu";
      game.sound_engine.gs("theme").setGain(30);
      game.sound_engine.gs("theme").rewind();
    }
  }
  
  void keyTyped() {
    state = "fade";
  }
}

class CreditsObject {
  String text;
  PVector location, olocation, elocation;
  float lerp_factor;
  float _delay;
  
  CreditsObject(String _text, float __delay) {
    text = _text;
    _delay = __delay;
    float t1 = random(TWO_PI);
    float t2 = t1+PI+random(-PI, PI)/4;
    olocation = new PVector(cos(t1)*width, sin(t1)*height);
    elocation = new PVector(cos(t2)*width, sin(t2)*height);
  }
  
  void update() {
    if (_delay > 0) {
      _delay -= game.dt;
    } else {
      lerp_factor += game.dt*0.003;
      location = PVector.lerp(olocation, elocation, lerp_factor);
      text(text, location.x, location.y);
    }
  }
}