class Particle {
  PVector location, velocity;
  float rotation, rotation_increase, rotation_slow, radius, fuse, max_fuse, fade_index, slowing, bounce;
  String type;
  boolean is_dead, affected_by_gravity, tangeble;
  color theColor;
  
  Particle(float _x, float _y, PVector _velocity, float _radius, float _rotation, float _rotation_increase, float _rotation_slow, String _type, float _fuse, color _theColor, float _fade_index, float _slowing, boolean _affected_by_gravity, boolean _tangeble, float _bounce) {
    location = new PVector(_x, _y);
    velocity = new PVector(_velocity.x, _velocity.y);
    radius = _radius;
    rotation = _rotation;
    rotation_increase = _rotation_increase;
    rotation_slow = _rotation_slow;
    type = _type;
    max_fuse = _fuse;
    fuse = _fuse;
    theColor = _theColor;
    fade_index = _fade_index;
    slowing = _slowing;
    affected_by_gravity = _affected_by_gravity;
    tangeble = _tangeble;
    bounce = _bounce;
  }
  
  void update() {
    //location.add(velocity);
    
    if (affected_by_gravity) {
      velocity.y += game.player.g;
    }
    
    location.x += velocity.x;
    if (tangeble && collidesWithBlocks()) {
      location.x -= velocity.x;
      velocity.x *= -bounce;
    }
    
    location.y += velocity.y;
    if (tangeble && collidesWithBlocks()) {
      location.y -= velocity.y;
      velocity.y *= -bounce;
    }
    
    velocity.mult(slowing);
    rotation += rotation_increase;
    rotation_increase *= rotation_slow;
    
    if (fuse == 0) {
      is_dead = true;
    }
    fuse--;
    
    display();
  }
  
  void display() {
    float alpha = 255;
    if (fuse <= max_fuse * fade_index) {
      alpha = 255*fuse/(max_fuse*fade_index);
    }
    fill(theColor, alpha);
    if (type.equals("circle")) {
      ellipse(location.x, location.y, radius, radius);
    } else if (type.equals("rect")) {
      pushMatrix();
      translate(location.x, location.y);
      rotate(radians(rotation));
      rect(0, 0, radius, radius);
      popMatrix();
    } else if (type.equals("triangle")) {
      pushMatrix();
      translate(location.x, location.y);
      rotate(radians(rotation));
      triangle(-radius/2, -sqrt(3)*radius/6, radius/2, -sqrt(3)*radius/6, 0, sqrt(3)*radius/3);
      popMatrix();
    }
  }
  
  boolean collidesWithBlocks() {
    for (HashMap.Entry<PVector,Block> entry : game.blocks.entrySet()) {
      Block block = entry.getValue();
      if (abs(location.x-block.location.x) <= 150 && abs(location.y-block.location.y) <= 150 && block.getOutsideCollision(location.x, location.y, 1, 1)) {
        if (block.collidable) {
          return true;
        }
      }
    }
    return false;
  }
}

class ParticleSystem {
  PVector location, velocity;
  boolean is_dead, affected_by_gravity, tangeble;
  Particle[] particles;
  float cooldown, max_cooldown, bounce;
  
  ParticleSystem(float x, float y, Particle[] _particles, float _cooldown) {
    location = new PVector(x, y);
    velocity = new PVector(0, 0);
    particles = _particles;
    is_dead = false;
    affected_by_gravity = false;
    tangeble = false;
    bounce = 0;
    max_cooldown = _cooldown;
    cooldown = 0;
  }
  
  ParticleSystem(float x, float y, float vx, float vy, Particle[] _particles, float _cooldown) {
    location = new PVector(x, y);
    velocity = new PVector(vx, vy);
    particles = _particles;
    is_dead = false;
    affected_by_gravity = false;
    tangeble = false;
    bounce = 0;
    max_cooldown = _cooldown;
    cooldown = 0;
  }
  
  ParticleSystem(float x, float y, float vx, float vy, Particle[] _particles, float _cooldown, boolean _affected_by_gravity, boolean _tangeble) {
    location = new PVector(x, y);
    velocity = new PVector(vx, vy);
    particles = _particles;
    is_dead = false;
    affected_by_gravity = _affected_by_gravity;
    tangeble = _tangeble;
    bounce = 0;
    max_cooldown = _cooldown;
    cooldown = 0;
  }
  
  void update() {
    move();
    spawn_particles();
  }
  
  void move() {
    //if (affected_by_gravity) {
    //  velocity.y += game.player.g;
    //}
    
    //location.x += velocity.x;
    //if (tangeble && collidesWithBlocks()) {
    //  location.x -= velocity.x;
    //  velocity.x *= -bounce;
    //}
    
    //location.y += velocity.y;
    //if (tangeble && collidesWithBlocks()) {
    //  location.y -= velocity.y;
    //  velocity.y *= -bounce;
    //}
  }
  
  void spawn_particles() {
    if (cooldown <= 0) {
      Particle particle = particles[(int)random(particles.length)];
      game.particles.add(new Particle(location.x, location.y, new PVector(random(-5, 5), random(-20, 5)), particle.radius, particle.rotation, particle.rotation_increase, particle.rotation_slow, particle.type, particle.fuse, particle.theColor, particle.fade_index, particle.slowing, particle.affected_by_gravity, particle.tangeble, particle.bounce));
      cooldown = max_cooldown;
    } else {
      cooldown -= game.dt;
    }
  }
  
  boolean collidesWithBlocks() {
    for (HashMap.Entry<PVector,Block> entry : game.blocks.entrySet()) {
      Block block = entry.getValue();
      if (abs(location.x-block.location.x) <= 150 && abs(location.y-block.location.y) <= 150 && block.getOutsideCollision(location.x, location.y, 1, 1)) {
        if (block.collidable) {
          return true;
        }
      }
    }
    return false;
  }
}