class Player {
  PVector location, velocity, rectSize, otherRectSize, checkpoint;
  String name;
  Animation animation, duck_animation;
  boolean w_pressed, a_pressed, d_pressed, ducking, is_dead, held_s, display_sideways, has_cannon, has_fuel;
  float g, jump, x_speed, cannon_x, cannon_y, cannon_cooldown, max_cannon_cooldown, falling_for;
  int cant_move_until, last_moved_by_block;
  
  Player(float x, float y) {
    location = new PVector(x, y);
    velocity = new PVector(0, 0);
    checkpoint = new PVector(x, y);
    name = "player_name";
    g = 1;
    jump = -18;
    x_speed = 6;
    display_sideways = false;
    falling_for = 0;
    is_dead = false;
    
    ArrayList<String> animation_phases = new ArrayList<String>();
    boolean out_of_imgs = false;
    for (int i = 1; !out_of_imgs; i++) {
      if (gt("player_"+i+".png") != null) {
        animation_phases.add("player_"+i+".png");
      } else {
        out_of_imgs = true;
      }
    }
    animation = new Animation(animation_phases, 100);
    
    ArrayList<String> duck_animation_phases = new ArrayList<String>();
    out_of_imgs = false;
    for (int i = 1; !out_of_imgs; i++) {
      if (gt("duck_"+i+".png") != null) {
        duck_animation_phases.add("duck_"+i+".png");
      } else {
        out_of_imgs = true;
      }
    }
    duck_animation = new Animation(duck_animation_phases, 200);
    
    rectSize = new PVector(gt(animation_phases.get(0)).width, gt(animation_phases.get(0)).height);
    otherRectSize = new PVector(gt(duck_animation_phases.get(0)).width, gt(duck_animation_phases.get(0)).height);
    
    has_fuel = false;
    
    has_cannon = false;
    cannon_cooldown = 0;
    max_cannon_cooldown = 15;
    cannon_x = -rectSize.x/4;
    cannon_y = -rectSize.y/3.5;
    
    cant_move_until = 0;
    
    last_moved_by_block = millis();
  }
  
  void update() {
    move();
    display();
  }
  
  void move() {
    if (cant_move_until > millis()) {
      return;
    }
    location.x += velocity.x*game.dt;
    if (collidesWithBlocks()) {
      location.x -= velocity.x*game.dt;
      velocity.x = 0;
    }
    velocity.x *= 0.9;
    
    float total_x_offset = 0;
    if (a_pressed) {
      total_x_offset -= x_speed;
    }
    if (d_pressed) {
      total_x_offset += x_speed;
    }
    
    location.x += total_x_offset*game.dt;
    if (collidesWithBlocks()) {
      location.x -= total_x_offset*game.dt;
    }
    
    velocity.y += g*game.dt;
    location.y += velocity.y*game.dt;
    if (collidesWithBlocks()) {
      location.y -= velocity.y*game.dt;
      velocity.y = 0;
      falling_for = 0;
    }
    
    if (velocity.y > 0) {
      falling_for += game.dt;
    } else {
      falling_for = 0;
    }
    
    if (w_pressed && velocity.y == 0) {
      location.y += g*game.dt;
      if (collidesWithBlocks()) {
        velocity.y = jump;
      }
      location.y -= g*game.dt;
    }
    
    if (has_cannon) {
      if (cannon_cooldown <= 0 && mousePressed) {
        if (mouseButton == LEFT) {
          cannon_cooldown = max_cannon_cooldown;
          PVector rotation_vector = new PVector(mouseX-width/2, mouseY-height/2).setMag(7);
          game.bullets.add(new PlayerBullet(location.x, location.y, rotation_vector.x, rotation_vector.y, random(TWO_PI), "middle_finger.png", 0.25));
          velocity.add(rotation_vector.mult(-2.5));
        } else if (mouseButton == RIGHT) {
          cannon_cooldown = max_cannon_cooldown*4;
          for (int i = -6; i <= 6; i++) {
            PVector rotation_vector = new PVector(mouseX-width/2, mouseY-height/2).rotate(radians(i*5)).setMag(7);
            game.bullets.add(new PlayerBullet(location.x, location.y, rotation_vector.x, rotation_vector.y, random(TWO_PI), "middle_finger.png", 0.25));
          }
          velocity.add(new PVector(mouseX-width/2, mouseY-height/2).setMag(7).mult(-4.5));
        }
      } else {
        cannon_cooldown -= game.dt;
      }
    }
    
    //if (location.y >= 1000) {
    if (falling_for > 40) {
      die("fall");
    }
  }
  
  void display() {
    pushMatrix();
    if (display_sideways) {
      translate(location.x, location.y-rectSize.x/2+rectSize.y/2);
      rotate(-PI/2);
    } else {
      translate(location.x, location.y);
    }
    if (!ducking) {
      image(animation.ci(), 0, 0);
    } else {
      image(duck_animation.ci(), 0, 0);
    }
    if (has_cannon || has_fuel) {
      imageMode(CORNER);
      PVector rotation_vector = new PVector(mouseX-width/2, mouseY-height/2).rotate(radians(12));
      rotate(rotation_vector.heading());
      if (has_cannon) {
        image(gt("cannon.png"), cannon_x, cannon_y);
      } else if (has_fuel) {
        image(gt("barrel.png"), cannon_x, cannon_y);
      }
      imageMode(CENTER);
    }
    popMatrix();
  }
  
  void die(String reason) {
    //println(reason);
    is_dead = true;
    animation.stopped = true;
    duck_animation.stopped = true;
    game.state = "player_dead";
    AudioPlayer pds = game.sound_engine.gs("lose_"+(int)random(1, 7));
    game.max_player_dead_counter = pds.length()/50.0*3.0;
    game.fail_counter++;
    pds.play();
  }
  
  boolean collidesWithBlocks() {
    return getCollisionBlock()!=null;
  }
  
  //Block getCollisionBlock() {
  //  for (HashMap.Entry<PVector,Block> entry : game.blocks.entrySet()) {
  //    Block block = entry.getValue();
  //    if (abs(location.x-block.location.x) <= 150 && abs(location.y-block.location.y) <= 150 && block.getPlayerSideCollision()) {
  //      block.onCollide();
  //      if (block.collidable) {
  //        return block;
  //      }
  //    }
  //  }
  //  return null;
  //}
  
  Block getCollisionBlock() {
    Block collision_block = null;
    for (HashMap.Entry<PVector,Block> entry : game.blocks.entrySet()) {
      Block block = entry.getValue();
      if (abs(location.x-block.location.x) <= 150 && abs(location.y-block.location.y) <= 150 && block.getPlayerSideCollision()) {
        block.onCollide();
        if (collision_block == null && block.collidable) {
          collision_block = block;
        }
      }
    }
    return collision_block;
  }
  
  void keyPressed() {
    if (key == 'w' || key == 'W') {
      w_pressed = true;
    }
    if (key == 'a' || key == 'A') {
      a_pressed = true;
    }
    if (key == 'd' || key == 'd') {
      d_pressed = true;
    }
    
    if (!ducking && (key == 's' || key == 'S')) {
      ducking = true;
      //location.y += rectSize.y-otherRectSize.y/2;
      PVector tempRectSize = new PVector(rectSize.x, rectSize.y);
      rectSize = new PVector(otherRectSize.x, otherRectSize.y);
      otherRectSize = new PVector(tempRectSize.x, tempRectSize.y);
      game.sound_engine.gs("duck").play();
    }
  }
  
  void keyReleased() {
    if (key == 'w' || key == 'W') {
      w_pressed = false;
    }
    if (key == 'a' || key == 'A') {
      a_pressed = false;
    }
    if (key == 'd' || key == 'd') {
      d_pressed = false;
    }
    
    if (held_s || (ducking && (key == 's' || key == 'S'))) {
      ducking = false;
      PVector tempRectSize = new PVector(rectSize.x, rectSize.y);
      rectSize = new PVector(otherRectSize.x, otherRectSize.y);
      otherRectSize = new PVector(tempRectSize.x, tempRectSize.y);
      location.y -= rectSize.y-otherRectSize.y/2;
      held_s = false;
      if (collidesWithBlocks()) {
        ducking = true;
        location.y += rectSize.y-otherRectSize.y/2;
        tempRectSize = new PVector(rectSize.x, rectSize.y);
        rectSize = new PVector(otherRectSize.x, otherRectSize.y);
        otherRectSize = new PVector(tempRectSize.x, tempRectSize.y);
        held_s = true;
        return;
      }
    }
    
    //debug
    if (key == 't') {
      location = new PVector(location.x+mouseX-width/2, location.y+mouseY-height/2);
    }
    if (key == 'c') {
      checkpoint = location.copy();
    }
  }
}